"""
Homomorphisms of abelian groups

.. TODO::

    - there must be a homspace first

    - there should be hom and Hom methods in abelian group

AUTHORS:

- David Joyner (2006-03-03): initial version
"""
# ****************************************************************************
#   Copyright (C) 2006 David Joyner and William Stein <wstein@gmail.com>
#
#   Distributed under the terms of the GNU General Public License (GPL)
#
#                    https://www.gnu.org/licenses/
# ****************************************************************************

from sage.libs.gap.libgap import libgap
from sage.categories.morphism import Morphism
from sage.misc.misc_c import prod


class AbelianGroupMap(Morphism):
    """
    A set-theoretic map between AbelianGroups.
    """
    def __init__(self, parent) -> None:
        """
        The Python constructor.
        """
        Morphism.__init__(self, parent)

    def _repr_type(self) -> str:
        return "AbelianGroup"


class AbelianGroupMorphism(Morphism):
    """
    Some python code for wrapping GAP's ``GroupHomomorphismByImages``
    function for abelian groups. Returns "fail" if ``gens`` does not
    generate ``self`` or if the map does not extend to a group
    homomorphism, ``self`` - ``other``.

    EXAMPLES::

        sage: G = AbelianGroup(3,[2,3,4],names='abc'); G
        Multiplicative Abelian group isomorphic to C2 x C3 x C4
        sage: a,b,c = G.gens()
        sage: H = AbelianGroup(2,[2,3],names='xy'); H
        Multiplicative Abelian group isomorphic to C2 x C3
        sage: x,y = H.gens()

        sage: from sage.groups.abelian_gps.abelian_group_morphism import AbelianGroupMorphism
        sage: phi = AbelianGroupMorphism(H,G,[x,y],[a,b])  # optional - gap_package_polycyclic

    TESTS::

        sage: G.<x,y> = AbelianGroup(2,[2,3])
        sage: H.<a,b,c> = AbelianGroup(3,[2,3,4])
        sage: phi = AbelianGroupMorphism(G,H,[x,y],[a,b])  # optional - gap_package_polycyclic
        sage: Hom(G,H) == phi.parent()                     # optional - gap_package_polycyclic
        True

    AUTHORS:

    - David Joyner (2006-02)
    """

#    There is a homomorphism from H to G but not from G to H:
#
#        sage: phi = AbelianGroupMorphism_im_gens(G,H,[a*b,a*c],[x,y])
#------------------------------------------------------------
#Traceback (most recent call last):
#  File "<ipython console>", line 1, in ?
#  File ".abeliangp_hom.sage.py", line 737, in __init__
#    raise TypeError("the orders of the corresponding elements in %s, %s must be equal" % (genss,imgss))
#TypeError: the orders of the corresponding elements in [a*b, a*c], [x, y] must be equal
#
#        sage: phi = AbelianGroupMorphism_im_gens(G,H,[a*b,(a*c)^2],[x*y,y])
#------------------------------------------------------------
#Traceback (most recent call last):
#  File "<ipython console>", line 1, in ?
#  File ".abeliangp_hom.sage.py", line 730, in __init__
#    raise TypeError("the list %s must generate G" % genss)
#TypeError: the list [a*b, c^2] must generate G

    def __init__(self, G, H, genss, imgss):
        from sage.categories.homset import Hom
        Morphism.__init__(self, Hom(G, H))
        if len(genss) != len(imgss):
            raise TypeError("the lengths of %s, %s must be equal" % (genss, imgss))
        self._domain = G
        self._codomain = H
        if not G.is_abelian():
            raise TypeError("the groups must be abelian groups")
        if not H.is_abelian():
            raise TypeError("the groups must be abelian groups")
        G_domain = G.subgroup(genss)
        if G_domain.order() != G.order():
            raise TypeError("the list %s must generate G" % genss)
        # self.domain_invs = G.gens_orders()
        # self.codomaininvs = H.gens_orders()
        self.domaingens = genss
        self.codomaingens = imgss
        for i in range(len(self.domaingens)):
            if (self.domaingens[i]).order() != (self.codomaingens[i]).order():
                raise TypeError("the orders of the corresponding elements in %s, %s must be equal" % (genss, imgss))

    def _libgap_(self):
        """
        Only works for finite groups.

        EXAMPLES::

            sage: G = AbelianGroup(3,[2,3,4],names='abc'); G
            Multiplicative Abelian group isomorphic to C2 x C3 x C4
            sage: a,b,c = G.gens()
            sage: H = AbelianGroup(2,[2,3],names='xy'); H
            Multiplicative Abelian group isomorphic to C2 x C3
            sage: x,y = H.gens()
            sage: phi = AbelianGroupMorphism(H,G,[x,y],[a,b])  # optional - gap_package_polycyclic
            sage: libgap(phi)                                  # optional - gap_package_polycyclic
            [ f1, f2 ] -> [ f1, f2 ]
            sage: phi = AbelianGroupMorphism(H,G,[x,y],[a*c**2,b])  # optional - gap_package_polycyclic
            sage: libgap(phi)                                       # optional - gap_package_polycyclic
            [ f1, f2 ] -> [ f1*f4, f2 ]
        """
        G = libgap(self.domain())
        H = libgap(self.codomain())
        in_G = [libgap(g) for g in self.domaingens]
        in_H = [libgap(h) for h in self.codomaingens]
        return G.GroupHomomorphismByImages(H, in_G, in_H)

    def _repr_type(self):
        return "AbelianGroup"

    def kernel(self):
        """
        Only works for finite groups.

        .. TODO::

            not done yet; returns a gap object but should return a Sage group.

        EXAMPLES::

            sage: H = AbelianGroup(3,[2,3,4],names='abc'); H
            Multiplicative Abelian group isomorphic to C2 x C3 x C4
            sage: a,b,c = H.gens()
            sage: G = AbelianGroup(2,[2,3],names='xy'); G
            Multiplicative Abelian group isomorphic to C2 x C3
            sage: x,y = G.gens()
            sage: phi = AbelianGroupMorphism(G,H,[x,y],[a,b])  # optional - gap_package_polycyclic
            sage: phi.kernel()                                 # optional - gap_package_polycyclic
            Group([  ])

            sage: H = AbelianGroup(3,[2,2,2],names='abc')
            sage: a,b,c = H.gens()
            sage: G = AbelianGroup(2,[2,2],names='x')
            sage: x,y = G.gens()
            sage: phi = AbelianGroupMorphism(G,H,[x,y],[a,a])  # optional - gap_package_polycyclic
            sage: phi.kernel()                                 # optional - gap_package_polycyclic
            Group([ f1*f2 ])
        """
        return libgap(self).Kernel()

    def image(self, S):
        """
        Return the image of the subgroup ``S`` by the morphism.

        This only works for finite groups.

        INPUT:

        - ``S`` -- a subgroup of the domain group ``G``

        EXAMPLES::

            sage: G = AbelianGroup(2,[2,3],names='xy')
            sage: x,y = G.gens()
            sage: subG = G.subgroup([x])                       # optional - gap_package_polycyclic
            sage: H = AbelianGroup(3,[2,3,4],names='abc')
            sage: a,b,c = H.gens()
            sage: phi = AbelianGroupMorphism(G,H,[x,y],[a,b])  # optional - gap_package_polycyclic
            sage: phi.image(subG)                              # optional - gap_package_polycyclic
            Multiplicative Abelian subgroup isomorphic to C2 generated by {a}
        """
        return self.codomain().subgroup([self(g) for g in S.gens()])

    def _call_(self, g):
        """
        Some python code for wrapping GAP's Images function but only for
        permutation groups. Returns an error if g is not in G.

        EXAMPLES::

            sage: H = AbelianGroup(3, [2,3,4], names='abc')
            sage: a,b,c = H.gens()
            sage: G = AbelianGroup(2, [2,3], names='xy')
            sage: x,y = G.gens()
            sage: phi = AbelianGroupMorphism(G,H,[x,y],[a,b])  # optional - gap_package_polycyclic
            sage: phi(y*x)                                     # optional - gap_package_polycyclic
            a*b
            sage: phi(y^2)                                     # optional - gap_package_polycyclic
            b^2
        """
        # g.word_problem is faster in general than word_problem(g)
        gens = self.codomaingens
        return prod(gens[(self.domaingens).index(wi[0])]**wi[1]
                    for wi in g.word_problem(self.domaingens))



# sage.doctest: needs sage.combinat sage.modules
r"""
Kleshchev partitions
====================

A partition (tuple) `\mu` is Kleshchev if it can be recursively
obtained by adding a sequence of good nodes to the empty
:class:`PartitionTuple` of the same :meth:`~PartitionTuple.level`
and *multicharge*. In this way, the set of Kleshchev multipartitions becomes
a realization of a Kashiwara crystal :mod:`sage.combinat.crystals.crystals`
for a irreducible integral highest weight representation of
`U_q(\widehat{\mathfrak{sl}}_e)`.

The Kleshchev multipartitions first appeared in the work of Ariki and Mathas
[AM2000]_ where it was shown that they index the irreducible representations
of the cyclotomic Hecke algebras of type `A` [AK1994]_. Soon afterwards Ariki
[Ariki2001]_ showed that the set of Kleshchev multipartitions naturally label
the irreducible representations of these algebras.  As a far reaching
generalization of these ideas the Ariki-Brundan-Kleshchev categorification
theorem [Ariki1996]_ [BK2009]_ says that these algebras categorify the
irreducible integral highest weight representations of the quantum group
`U_q(\widehat{\mathfrak{sl}}_e)` of the affine special linear group. Under
this categorification, `q` corresponds to the grading shift on the
cyclotomic Hecke algebras, where the grading from the Brundan-Kleshchev
graded isomorphism theorem to the *KLR algebras* of type `A` [BK2009]_.

The group algebras of the symmetric group in characteristic `p` are an
important special case of the cyclotomic Hecke algebras of type `A`.
In this case, depending on your prefer convention, the set of Kleshchev
partitions is the set of *`p`-regular* or *`p`-restricted*
:class:`~sage.combinat.partition.Partitions`. In this case, Kleshchev
[Kle1995]_ proved that the *modular branching rules* were given by adding
and removing *good nodes*; see :meth:`~KleshchevPartition.good_cells`.
Lascoux, Leclerc and Thibon [LLT1996]_ noticed that Kleshchev's branching
rules coincided with Kashiwara's crystal operators for the fundamental
representation of `L(\Lambda_0)` of `U_q(\widehat{\mathfrak{sl}}_p)`
and their celebrated *LLT conjecture* said that decomposition matrices of
the :class:`sage.algebras.iwahori_hecke_algebra.IwahoriHeckeAlgebra` of
the symmetric group should be computable using the canonical basis of
`L(\Lambda_0)`. This was proved and generalised to all cyclotomic Hecke
algebras of type `A` by Ariki [Ariki1996]_ and then further generalized
to the graded setting by Brundan and Kleshchev [BK2009]_.

The main class for accessing Kleshchev partition (tuples) is
:class:`~KleshchevPartitions`. Unfortunately, just as with the
symmetric group, different authors use different conventions when
defining Kleshchev partitions, which depends on whether you read
components from left to right, or right to left, and whether you
read the nodes in the partition in each component from top to bottom
or bottom to top. The :class:`~KleshchevPartitions` class supports
these four different conventions::

    sage: KleshchevPartitions(2, [0,0], size=2, convention='left regular')[:]
    [([1], [1]), ([2], [])]
    sage: KleshchevPartitions(2, [0,0], size=2, convention='left restricted')[:]
    [([1], [1]), ([], [1, 1])]
    sage: KleshchevPartitions(2, [0,0], size=2, convention='right regular')[:]
    [([1], [1]), ([], [2])]
    sage: KleshchevPartitions(2, [0,0], size=2, convention='right restricted')[:]
    [([1], [1]), ([1, 1], [])]

By default, the ``left restricted`` convention is used. As a shorthand,
``LG``, ``LS``, ``RG`` and ``RS``, respectively, can be used to specify the
``convention`` With the ``left`` convention the partition tuples should be
ordered with the most dominant partitions in the partition tuple on the left
and with the ``right`` convention the most dominant partition is on the right.

The :class:`~KleshchevPartitions` class can automatically convert between
these four different conventions::

    sage: KPlg = KleshchevPartitions(2, [0,0], size=2, convention='left regular')
    sage: KPls = KleshchevPartitions(2, [0,0], size=2, convention='left restricted')
    sage: [KPlg(mu) for mu in KPls] # indirect doc test
    [([1], [1]), ([2], [])]

AUTHORS:

- Andrew Mathas and Travis Scrimshaw (2018-05-1): Initial version
"""

from .partition import Partition, Partitions
from .partition_tuple import PartitionTuple, PartitionTuples

from sage.categories.finite_enumerated_sets import FiniteEnumeratedSets
from sage.categories.infinite_enumerated_sets import InfiniteEnumeratedSets
from sage.misc.lazy_attribute import lazy_attribute
from sage.rings.finite_rings.integer_mod_ring import IntegerModRing
from sage.rings.integer_ring import ZZ
from sage.rings.semirings.non_negative_integer_semiring import NN
from sage.cpython.getattr import getattr_from_other_class

from collections import defaultdict

#--------------------------------------------------
# Kleshchev partition - element classes
#--------------------------------------------------


class KleshchevPartition(Partition):
    r"""
    Abstract base class for Kleshchev partitions. See
    :class:`~KleshchevPartitions`.
    """

    def conormal_cells(self, i=None):
        r"""
        Return a dictionary of the cells of ``self`` which are conormal.

        Following [Kle1995]_, the *conormal* cells are computed by
        reading up (or down) the rows of the partition and marking all
        of the addable and removable cells of `e`-residue `i` and then
        recursively removing all adjacent pairs of removable and addable
        cells (in that order) from this list. The addable `i`-cells that
        remain at the end of the this process are the conormal `i`-cells.

        When computing conormal cells you can either read the cells in order
        from top to bottom (this corresponds to labeling the simple modules
        of the symmetric group by regular partitions) or from bottom to top
        (corresponding to labeling the simples by restricted partitions).
        By default we read down the partition but this can be changed by
        setting ``convention = 'RS'``.

        INPUT:

        - ``i`` -- (optional) a residue

        OUTPUT:

        If no residue ``i`` is specified then a dictionary of conormal cells
        is returned, which gives the conormal cells for ``0 <= i < e``.

        EXAMPLES::

            sage: KP = KleshchevPartitions(3, convention='regular')
            sage: KP([5,4,4,3,2]).conormal_cells()
            {0: [(1, 4)], 1: [(5, 0), (4, 2)]}
            sage: KP([5,4,4,3,2]).conormal_cells(0)
            [(1, 4)]
            sage: KP([5,4,4,3,2]).conormal_cells(1)
            [(5, 0), (4, 2)]
            sage: KP = KleshchevPartitions(3, convention='restricted')
            sage: KP([5,4,4,3,2]).conormal_cells()
            {0: [(1, 4), (3, 3)], 2: [(0, 5)]}
        """
        # We use a dictionary for the conormal nodes as the indexing set is Z when e=0
        conormals = defaultdict(list)   # the conormal cells of each residue
        carry = defaultdict(int)        # a tally of #(removable cells) - #(addable cells)

        # determine if we read up or down the partition
        KP = self.parent()
        rows = list(range(len(self)+1))
        if KP._convention[1] == 'G':
            rows.reverse()

        # work through the rows
        for row in rows:
            if row == len(self): # addable cell at bottom of partition
                res = KP._multicharge[0] - row
                if carry[res] == 0:
                    conormals[res].append((row, 0))
                else:
                    carry[res] += 1
            else:
                res = KP._multicharge[0] + self[row] - row - 1
                if row == len(self)-1 or self[row] > self[row+1]:  # removable cell
                    carry[res] -= 1
                if row == 0 or self[row-1] > self[row]:  # addable cell
                    if carry[res+1] >= 0:
                        conormals[res+1].append((row, self[row]))
                    else:
                        carry[res+1] += 1

        # finally return the result
        return dict(conormals) if i is None else conormals[i]

    def cogood_cells(self, i=None):
        r"""
        Return a list of the cells of ``self`` that are cogood.

        The cogood `i`-cell is the 'last' conormal `i`-cell. As with the
        conormal cells we can choose to read either up or down the partition as
        specified by :meth:`~KleshchevPartitions.convention`.

        INPUT:

        - ``i`` -- (optional) a residue

        OUTPUT:

        If no residue ``i`` is specified then a dictionary of cogood cells
        is returned, which gives the cogood cells for ``0 <= i < e``.

        EXAMPLES::

            sage: KP = KleshchevPartitions(3, convention='regular')
            sage: KP([5,4,4,3,2]).cogood_cells()
            {0: (1, 4), 1: (4, 2)}
            sage: KP([5,4,4,3,2]).cogood_cells(0)
            (1, 4)
            sage: KP([5,4,4,3,2]).cogood_cells(1)
            (4, 2)
            sage: KP = KleshchevPartitions(4, convention='restricted')
            sage: KP([5,4,4,3,2]).cogood_cells()
            {1: (0, 5), 2: (4, 2), 3: (1, 4)}
            sage: KP([5,4,4,3,2]).cogood_cells(0)
            sage: KP([5,4,4,3,2]).cogood_cells(2)
            (4, 2)
        """
        conormal_cells = self.conormal_cells(i)
        if i is None:
            return {i: conormal_cells[i][-1] for i in conormal_cells}
        if not conormal_cells:
            return None

        return conormal_cells[-1]

    def normal_cells(self, i=None):
        r"""
        Return a dictionary of the cells of the partition that are normal.

        Following [Kle1995]_, the *normal* cells are computed by
        reading up (or down) the rows of the partition and marking all
        of the addable and removable cells of `e`-residue `i` and then
        recursively removing all adjacent pairs of removable and
        addable cells (in that order) from this list. The removable
        `i`-cells that remain at the end of the this process are the
        normal `i`-cells.

        When computing normal cells you can either read the cells in order
        from top to bottom (this corresponds to labeling the simple modules
        of the symmetric group by regular partitions) or from bottom to top
        (corresponding to labeling the simples by restricted partitions).
        By default we read down the partition but this can be changed by
        setting ``convention = 'RS'``.

        INPUT:

        - ``i`` -- (optional) a residue

        OUTPUT:

        If no residue ``i`` is specified then a dictionary of normal cells
        is returned, which gives the normal cells for ``0 <= i < e``.

        EXAMPLES::

            sage: KP = KleshchevPartitions(3, convention='regular')
            sage: KP([5,4,4,3,2]).normal_cells()
            {1: [(2, 3), (0, 4)]}
            sage: KP([5,4,4,3,2]).normal_cells(1)
            [(2, 3), (0, 4)]
            sage: KP = KleshchevPartitions(3, convention='restricted')
            sage: KP([5,4,4,3,2]).normal_cells()
            {0: [(4, 1)], 2: [(3, 2)]}
            sage: KP([5,4,4,3,2]).normal_cells(2)
            [(3, 2)]
        """
        # We use a dictionary for the normal nodes as the indexing set is Z when e=0
        normals = defaultdict(list)     # the normal cells of each residue
        carry = defaultdict(int)        # a tally of #(removable cells)-#(addable cells)

        # determine if we read up or down the partition
        KP = self.parent()
        rows = list(range(len(self)+1))
        if KP._convention[1] == 'S':
            rows.reverse()

        # work through the rows
        for row in rows:
            if row == len(self): # addable cell at bottom of partition
                carry[KP._multicharge[0]-row] += 1
            else:
                res = KP._multicharge[0] + self[row] - row - 1
                if row == len(self) - 1 or self[row] > self[row+1]: # removable cell
                    if carry[res] == 0:
                        normals[res].insert(0, (row, self[row]-1))
                    else:
                        carry[res] -= 1
                if row == 0 or self[row-1] > self[row]:              # addable cell
                    carry[res+1] += 1

        # finally return the result
        return dict(normals) if i is None else normals[i]

    def good_cells(self, i=None):
        """
        Return a list of the cells of ``self`` that are good.

        The good `i`-cell is the 'first' normal `i`-cell. As with the normal
        cells we can choose to read either up or down the partition as
        specified by :meth:`~KleshchevPartitions.convention`.

        INPUT:

        - ``i`` -- (optional) a residue

        OUTPUT:

        If no residue ``i`` is specified then a dictionary of good cells
        is returned, which gives the good cells for ``0 <= i < e``.

        EXAMPLES::

            sage: KP3 = KleshchevPartitions(3, convention='regular')
            sage: KP3([5,4,4,3,2]).good_cells()
            {1: (2, 3)}
            sage: KP3([5,4,4,3,2]).good_cells(1)
            (2, 3)
            sage: KP4 = KleshchevPartitions(4, convention='restricted')
            sage: KP4([5,4,4,3,2]).good_cells()
            {1: (2, 3)}
            sage: KP4([5,4,4,3,2]).good_cells(0)
            sage: KP4([5,4,4,3,2]).good_cells(1)
            (2, 3)
        """
        normal_cells = self.normal_cells(i)
        if i is None:
            return {j: normal_cells[j][0] for j in normal_cells}
        if not normal_cells:
            return None

        return normal_cells[0]

    def good_residue_sequence(self):
        """
        Return a sequence of good nodes from the empty partition
        to ``self``, or ``None`` if no such sequence exists.

        EXAMPLES::

            sage: KP = KleshchevPartitions(3, convention='regular')
            sage: KP([5,4,4,3,2]).good_residue_sequence()
            [0, 2, 1, 1, 0, 2, 0, 2, 1, 1, 0, 2, 0, 2, 2, 0, 1, 1]
            sage: KP = KleshchevPartitions(3, convention='restricted')
            sage: KP([5,4,4,3,2]).good_residue_sequence()
            [0, 1, 2, 2, 0, 1, 0, 2, 1, 2, 0, 1, 0, 2, 1, 2, 1, 0]
        """
        if not self:
            return []

        good_cells = self.good_cells()
        assert good_cells

        res = sorted(good_cells)[0]
        r, c = good_cells[res]
        good_seq = type(self)(self.parent(), self.remove_cell(r,c)).good_residue_sequence()
        good_seq.append(self.parent()._index_set(res))
        return good_seq

    def good_cell_sequence(self):
        """
        Return a sequence of good nodes from the empty partition
        to ``self``, or ``None`` if no such sequence exists.

        EXAMPLES::

            sage: KP = KleshchevPartitions(3, convention='regular')
            sage: KP([5,4,4,3,2]).good_cell_sequence()
            [(0, 0), (1, 0), (0, 1), (2, 0), (1, 1), (0, 2),
             (3, 0), (2, 1), (1, 2), (3, 1), (0, 3), (1, 3),
             (2, 2), (3, 2), (4, 0), (4, 1), (0, 4), (2, 3)]
            sage: KP = KleshchevPartitions(3, convention='restricted')
            sage: KP([5,4,4,3,2]).good_cell_sequence()
            [(0, 0), (0, 1), (1, 0), (0, 2), (1, 1), (2, 0),
             (0, 3), (2, 1), (1, 2), (1, 3), (3, 0), (3, 1),
             (2, 2), (4, 0), (2, 3), (3, 2), (0, 4), (4, 1)]
        """
        if not self:
            return []

        good_cells = self.good_cells()
        assert good_cells

        cell = good_cells[sorted(good_cells)[0]]
        good_seq = type(self)(self.parent(), self.remove_cell(*cell)).good_cell_sequence()
        good_seq.append(cell)
        return good_seq

    def mullineux_conjugate(self):
        r"""
        Return the partition tuple that is the Mullineux conjugate of ``self``.

        It follows from results in [BK2009]_, [Mat2015]_ that if `\nu` is the
        Mullineux conjugate of the Kleshchev partition tuple `\mu` then the
        simple module `D^\nu =(D^\mu)^{\text{sgn}}` is obtained from `D^\mu`
        by twisting by the `\text{sgn}`-automorphism with is the
        Iwahori-Hecke algebra analogue of tensoring with the one
        dimensional sign representation.

        EXAMPLES::

            sage: KP = KleshchevPartitions(3, convention='regular')
            sage: KP([5,4,4,3,2]).mullineux_conjugate()
            [9, 7, 1, 1]
            sage: KP = KleshchevPartitions(3, convention='restricted')
            sage: KP([5,4,4,3,2]).mullineux_conjugate()
            [3, 2, 2, 2, 2, 2, 2, 1, 1, 1]
            sage: KP = KleshchevPartitions(3, [2], convention='regular')
            sage: mc = KP([5,4,4,3,2]).mullineux_conjugate(); mc
            [9, 7, 1, 1]
            sage: mc.parent().multicharge()
            (1,)
            sage: KP = KleshchevPartitions(3, [2], convention='restricted')
            sage: mc = KP([5,4,4,3,2]).mullineux_conjugate(); mc
            [3, 2, 2, 2, 2, 2, 2, 1, 1, 1]
            sage: mc.parent().multicharge()
            (1,)
        """
        P = self.parent()
        if not self:
            size = None
            if isinstance(P, KleshchevPartitions_size):
                size = P._size
            KP = KleshchevPartitions(P._e, [-c for c in P._multicharge],
                                     size=size, convention=P._convention)
            return KP.element_class(KP, [])

        good_cells = self.good_cells()
        assert good_cells

        r, c = sorted(good_cells.values())[0]
        # This is technically wrong when the parent has a fixed size because
        #   the resulting Kleshchev partition after removing a cell has abs
        #   smaller size. However, this is useful to avoid constructing
        #   transient parents.
        mu = P.element_class(P, self.remove_cell(r, c)).mullineux_conjugate()
        # add back on a cogood cell of residue -residue(k,r,c)
        KP = mu.parent()
        return KP.element_class(
            KP,
            mu.add_cell(*mu.cogood_cells(r-c-self.parent()._multicharge[0])))

    def is_regular(self) -> bool:
        r"""
        Return ``True`` if ``self`` is a `e`-regular partition tuple.

        A partition tuple is `e`-regular if we can get to the empty partition
        tuple by successively removing a sequence of good cells in the down
        direction. Equivalently, all partitions are `0`-regular and if `e > 0`
        then a partition is `e`-regular if no `e` nonzero parts of ``self``
        are equal.

        EXAMPLES::

            sage: KP = KleshchevPartitions(2)
            sage: KP([2,1,1]).is_regular()
            False
            sage: KP = KleshchevPartitions(3)
            sage: KP([2,1,1]).is_regular()
            True
            sage: KP([]).is_regular()
            True
        """
        if self.size() == 0 or self.parent()._e == 0:
            return True
        KP = self.parent()
        return super().is_regular(KP._e, KP._multicharge)

    def is_restricted(self) -> bool:
        r"""
        Return ``True`` if ``self`` is an `e`-restricted partition tuple.

        A partition tuple is `e`-restricted if we can get to the empty
        partition tuple by successively removing a sequence of good cells in
        the up direction. Equivalently, all partitions are `0`-restricted and
        if `e > 0` then a partition is `e`-restricted if the difference of
        successive parts of ``self`` are always strictly less than `e`.

        EXAMPLES::

            sage: KP = KleshchevPartitions(2, convention='regular')
            sage: KP([3,1]).is_restricted()
            False
            sage: KP = KleshchevPartitions(3, convention='regular')
            sage: KP([3,1]).is_restricted()
            True
            sage: KP([]).is_restricted()
            True
        """
        if self.size() == 0 or self.parent()._e == 0:
            return True
        KP = self.parent()
        return super().is_restricted(KP._e, KP._multicharge)


class KleshchevPartitionTuple(PartitionTuple):
    r"""
    Abstract base class for Kleshchev partition tuples. See
    :class:`~KleshchevPartitions`.
    """

    def conormal_cells(self, i=None):
        r"""
        Return a dictionary of the cells of the partition that are conormal.

        Following [Kle1995]_, the *conormal* cells are computed by
        reading up (or down) the rows of the partition and marking all
        of the addable and removable cells of `e`-residue `i` and then
        recursively removing all adjacent pairs of removable and addable
        cells (in that order) from this list. The addable `i`-cells that
        remain at the end of the this process are the conormal `i`-cells.

        When computing conormal cells you can either read the cells in order
        from top to bottom (this corresponds to labeling the simple modules
        of the symmetric group by regular partitions) or from bottom to top
        (corresponding to labeling the simples by restricted partitions).
        By default we read down the partition but this can be changed by
        setting ``convention = 'RS'``.

        INPUT:

        - ``i`` -- (optional) a residue

        OUTPUT:

        If no residue ``i`` is specified then a dictionary of conormal cells
        is returned, which gives the conormal cells for ``0 <= i < e``.

        EXAMPLES::

            sage: KP = KleshchevPartitions(3, [0,1], convention="left regular")
            sage: KP([[4, 2], [5, 3, 1]]).conormal_cells()
            {0: [(1, 2, 1), (1, 1, 3), (1, 0, 5)],
            1: [(1, 3, 0), (0, 2, 0), (0, 1, 2), (0, 0, 4)]}
            sage: KP([[4, 2], [5, 3, 1]]).conormal_cells(1)
            [(1, 3, 0), (0, 2, 0), (0, 1, 2), (0, 0, 4)]
            sage: KP([[4, 2], [5, 3, 1]]).conormal_cells(2)
            []
            sage: KP = KleshchevPartitions(3, [0,1], convention="right restricted")
            sage: KP([[4, 2], [5, 3, 1]]).conormal_cells(0)
            [(1, 0, 5), (1, 1, 3), (1, 2, 1)]
        """
        # We use a dictionary for the conormal nodes as the indexing set is Z when e=0
        conormals = defaultdict(list)   # the conormal cells of each residue
        carry = defaultdict(int)        # a tally of #(removable cells)-#(addable cells)

        part_lens = [len(part) for part in self]  # so we don't repeatedly call these
        # the indices for the rows ending in addable nodes
        KP = self.parent()
        if KP._convention[0] == 'L':
            rows = [(k,r) for k,ell in enumerate(part_lens) for r in range(ell+1)]
        else:
            rows = [(k,r) for k,ell in reversed(list(enumerate(part_lens))) for r in range(ell+1)]
        if KP._convention[1] == 'G':
            rows.reverse()

        for row in rows:
            k,r = row
            if r == part_lens[k]: # addable cell at bottom of a component
                res = KP._multicharge[k] - r
                if carry[res] == 0:
                    conormals[res].append((k, r, 0))
                else:
                    carry[res] += 1
            else:
                part = self[k]
                res = KP._multicharge[k] + (part[r] - r - 1)
                if r == part_lens[k] - 1 or part[r] > part[r+1]: # removable cell
                    carry[res] -= 1
                if r == 0 or part[r-1] > part[r]:                # addable cell
                    if carry[res+1] == 0:
                        conormals[res+1].append((k, r, part[r]))
                    else:
                        carry[res+1] += 1

        # finally return the result
        if i is None:
            return dict(conormals)
        return conormals[i]

    def cogood_cells(self, i=None):
        r"""
        Return a list of the cells of the partition that are cogood.

        The cogood `i`-cell is the 'last' conormal `i`-cell. As with the
        conormal cells we can choose to read either up or down the partition
        as specified by :meth:`~KleshchevPartitions.convention`.

        INPUT:

        - ``i`` -- (optional) a residue

        OUTPUT:

        If no residue ``i`` is specified then a dictionary of cogood cells
        is returned, which gives the cogood cells for ``0 <= i < e``.

        EXAMPLES::

            sage: KP = KleshchevPartitions(3, [0,1])
            sage: pt = KP([[4, 2], [5, 3, 1]])
            sage: pt.cogood_cells()
            {0: (1, 2, 1), 1: (1, 3, 0)}
            sage: pt.cogood_cells(0)
            (1, 2, 1)
            sage: KP = KleshchevPartitions(4, [0,1], convention="left regular")
            sage: pt = KP([[5, 2, 2], [6, 1, 1]])
            sage: pt.cogood_cells()
            {1: (0, 0, 5), 2: (1, 3, 0)}
            sage: pt.cogood_cells(0) is None
            True
            sage: pt.cogood_cells(1) is None
            False
        """
        conormal_cells = self.conormal_cells(i)
        if i is None:
            return {j: conormal_cells[j][-1] for j in conormal_cells}
        if not conormal_cells:
            return None

        return conormal_cells[-1]

    def normal_cells(self, i=None):
        r"""
        Return a dictionary of the removable cells of the partition that
        are normal.

        Following [Kle1995]_, the *normal* cells are computed by
        reading up (or down) the rows of the partition and marking all
        of the addable and removable cells of `e`-residue `i` and then
        recursively removing all adjacent pairs of removable and
        addable cells (in that order) from this list. The removable
        `i`-cells that remain at the end of the this process are the
        normal `i`-cells.

        When computing normal cells you can either read the cells in order
        from top to bottom (this corresponds to labeling the simple modules
        of the symmetric group by regular partitions) or from bottom to top
        (corresponding to labeling the simples by restricted partitions).
        By default we read down the partition but this can be changed by
        setting ``convention = 'RS'``.

        INPUT:

        - ``i`` -- (optional) a residue

        OUTPUT:

        If no residue ``i`` is specified then a dictionary of normal cells
        is returned, which gives the normal cells for ``0 <= i < e``.

        EXAMPLES::

            sage: KP = KleshchevPartitions(3, [0,1], convention="left restricted")
            sage: KP([[4, 2], [5, 3, 1]]).normal_cells()
            {2: [(1, 0, 4), (1, 1, 2), (1, 2, 0)]}
            sage: KP([[4, 2], [5, 3, 1]]).normal_cells(1)
            []
            sage: KP = KleshchevPartitions(3, [0,1], convention="left regular")
            sage: KP([[4, 2], [5, 3, 1]]).normal_cells()
            {0: [(0, 1, 1), (0, 0, 3)], 2: [(1, 2, 0), (1, 1, 2), (1, 0, 4)]}
            sage: KP = KleshchevPartitions(3, [0,1], convention="right regular")
            sage: KP([[4, 2], [5, 3, 1]]).normal_cells()
            {2: [(1, 2, 0), (1, 1, 2), (1, 0, 4)]}
            sage: KP = KleshchevPartitions(3, [0,1], convention="right restricted")
            sage: KP([[4, 2], [5, 3, 1]]).normal_cells()
            {0: [(0, 0, 3), (0, 1, 1)], 2: [(1, 0, 4), (1, 1, 2), (1, 2, 0)]}
        """
        # We use a dictionary for the normal nodes as the indexing set is Z when e=0
        normals = defaultdict(list)     # the normal cells of each residue
        carry = defaultdict(int)        # a tally of #(removable cells)-#(addable cells)

        part_lens = [len(part) for part in self]  # so we don't repeatedly call these
        KP = self.parent()
        if KP._convention[0] == 'L':
            rows = [(k, r) for k, ell in enumerate(part_lens) for r in range(ell+1)]
        else:
            rows = [(k, r) for k, ell in reversed(list(enumerate(part_lens))) for r in range(ell+1)]
        if KP._convention[1] == 'S':
            rows.reverse()

        for row in rows:
            k, r = row
            if r == part_lens[k]:  # addable cell at bottom of a component
                carry[KP._multicharge[k]-r] += 1
            else:
                part = self[k]
                res = KP._multicharge[k] + (part[r] - r - 1)
                if r == part_lens[k]-1 or part[r] > part[r+1]:  # removable cell
                    if carry[res] == 0:
                        normals[res].insert(0, (k, r, part[r]-1))
                    else:
                        carry[res] -= 1
                if r == 0 or part[r-1] > part[r]:     # addable cell
                    carry[res+1] += 1

        # finally return the result
        if i is None:
            return dict(normals)    # change the defaultdict into a dict

        return normals[i]

    def good_cells(self, i=None):
        r"""
        Return a list of the cells of the partition tuple which are good.

        The good `i`-cell is the 'first' normal `i`-cell. As with the normal
        cells we can choose to read either up or down the partition as specified
        by :meth:`~KleshchevPartitions.convention`.

        INPUT:

        - ``i`` -- (optional) a residue

        OUTPUT:

        If no residue ``i`` is specified then a dictionary of good cells
        is returned, which gives the good cells for ``0 <= i < e``.

        EXAMPLES::

            sage: KP = KleshchevPartitions(3, [0,1])
            sage: pt = KP([[4, 2], [5, 3, 1]])
            sage: pt.good_cells()
            {2: (1, 0, 4)}
            sage: pt.good_cells(2)
            (1, 0, 4)
            sage: KP = KleshchevPartitions(4, [0,1], convention="left regular")
            sage: pt = KP([[5, 2, 2], [6, 2, 1]])
            sage: pt.good_cells()
            {0: (0, 0, 4), 2: (1, 0, 5), 3: (0, 2, 1)}
            sage: pt.good_cells(1) is None
            True
        """
        normal_cells = self.normal_cells(i)
        if i is None:
            return {j: normal_cells[j][0] for j in normal_cells}
        if not normal_cells:
            return None

        return normal_cells[0]

    def good_residue_sequence(self):
        """
        Return a sequence of good nodes from the empty partition to ``self``.

        EXAMPLES::

            sage: KP = KleshchevPartitions(3, [0,1])
            sage: KP([[4, 2], [5, 3, 1]]).good_residue_sequence()
            [0, 1, 2, 1, 2, 0, 1, 0, 2, 2, 0, 1, 0, 2, 2]
        """
        if self.size() == 0:
            return []
        good_cells = self.good_cells()
        assert good_cells

        res = sorted(good_cells.keys())[0]
        k, r, c = good_cells[res]
        good_seq = type(self)(self.parent(), self.remove_cell(k, r, c)).good_residue_sequence()
        good_seq.append(self.parent()._index_set(res))
        return good_seq

    def good_cell_sequence(self):
        """
        Return a sequence of good nodes from the empty partition to ``self``.

        EXAMPLES::

            sage: KP = KleshchevPartitions(3,[0,1])
            sage: KP([[4, 2], [5, 3, 1]]).good_cell_sequence()
            [(0, 0, 0), (1, 0, 0), (1, 0, 1), (0, 0, 1), (0, 1, 0),
             (1, 1, 0), (1, 1, 1), (1, 0, 2), (1, 2, 0), (0, 0, 2),
             (0, 1, 1), (1, 0, 3), (0, 0, 3), (1, 1, 2), (1, 0, 4)]
        """
        if self.size() == 0:
            return []
        good_cells = self.good_cells()
        assert good_cells

        cell = good_cells[sorted(good_cells)[0]]
        good_seq = type(self)(self.parent(), self.remove_cell(*cell)).good_cell_sequence()
        good_seq.append(cell)
        return good_seq

    def mullineux_conjugate(self):
        r"""
        Return the partition that is the Mullineux conjugate of ``self``.

        It follows from results in [Kle1996]_ [Bru1998]_ that if `\nu` is the
        Mullineux conjugate of the Kleshchev partition tuple `\mu` then the
        simple module `D^\nu =(D^\mu)^{\text{sgn}}` is obtained from `D^\mu`
        by twisting by the `\text{sgn}`-automorphism with is the Hecke algebra
        analogue of tensoring with the one dimensional sign representation.

        EXAMPLES::

            sage: KP = KleshchevPartitions(3, [0,1])
            sage: mc = KP([[4, 2], [5, 3, 1]]).mullineux_conjugate(); mc
            ([2, 2, 1, 1], [3, 2, 2, 1, 1])
            sage: mc.parent()
            Kleshchev partitions with e=3 and multicharge=(0,2)
        """
        P = self.parent()
        if self.size() == 0:
            size = None
            if isinstance(P, KleshchevPartitions_size):
                size = P._size
            KP = KleshchevPartitions(P._e, [-c for c in P._multicharge],
                                     size=size, convention=P._convention)
            return KP.element_class(KP, [[]]*P._level)

        good_cells = self.good_cells()
        assert good_cells

        k,r,c = sorted(good_cells.values())[0]
        # This is technically wrong when the parent has a fixed size because
        #   the resulting Kleshchev partition after removing a cell has abs
        #   smaller size. However, this is useful to avoid constructing
        #   transient parents.
        mu = P.element_class(P, self.remove_cell(k,r,c)).mullineux_conjugate()
        # add back on a cogood cell of residue -residue(k,r,c)
        KP = mu.parent()
        return KP.element_class(KP, mu.add_cell(*mu.cogood_cells( r-c-self.parent()._multicharge[k])))

    def is_regular(self) -> bool:
        r"""
        Return ``True`` if ``self`` is a `e`-regular partition tuple.

        A partition tuple is `e`-regular if we can get to the
        empty partition tuple by successively removing a sequence
        of good cells in the down direction.

        EXAMPLES::

            sage: KP = KleshchevPartitions(2, [0,2], convention="right restricted")
            sage: KP([[3,2,1], [2,1,1]]).is_regular()
            False
            sage: KP = KleshchevPartitions(4, [0,2], convention="right restricted")
            sage: KP([[3,2,1], [2,1,1]]).is_regular()
            True
            sage: KP([[], []]).is_regular()
            True
        """
        if self.size() == 0:
            return True
        KP = self.parent()
        return _is_regular(self.to_list(), KP._multicharge, KP._convention)

    def is_restricted(self) -> bool:
        r"""
        Return ``True`` if ``self`` is an `e`-restricted partition tuple.

        A partition tuple is `e`-restricted if we can get to the
        empty partition tuple by successively removing a sequence
        of good cells in the up direction.

        EXAMPLES::

            sage: KP = KleshchevPartitions(2, [0,2], convention="left regular")
            sage: KP([[3,2,1], [3,1]]).is_restricted()
            False
            sage: KP = KleshchevPartitions(3, [0,2], convention="left regular")
            sage: KP([[3,2,1], [3,1]]).is_restricted()
            True
            sage: KP([[], []]).is_restricted()
            True
        """
        if self.size() == 0:
            return True
        KP = self.parent()
        return _is_restricted(self.to_list(), KP._multicharge, KP._convention)


class KleshchevCrystalMixin:
    """
    Mixin class for the crystal structure of a Kleshchev partition.
    """

    def epsilon(self, i):
        r"""
        Return the Kashiwara crystal operator `\varepsilon_i` applied to ``self``.

        INPUT:

        - ``i`` -- an element of the index set

        EXAMPLES::

            sage: C = crystals.KleshchevPartitions(3, [0,2], convention="left regular")
            sage: x = C([[5,4,1],[3,2,1,1]])
            sage: [x.epsilon(i) for i in C.index_set()]
            [0, 3, 0]
        """
        return len(self.normal_cells(i))

    def phi(self, i):
        r"""
        Return the Kashiwara crystal operator `\varphi_i` applied to ``self``.

        INPUT:

        - ``i`` -- an element of the index set

        EXAMPLES::

            sage: C = crystals.KleshchevPartitions(3, [0,2], convention="left regular")
            sage: x = C([[5,4,1],[3,2,1,1]])
            sage: [x.phi(i) for i in C.index_set()]
            [3, 2, 0]
        """
        return len(self.conormal_cells(i))

    def Epsilon(self):
        r"""
        Return `\varepsilon` of ``self``.

        EXAMPLES::

            sage: C = crystals.KleshchevPartitions(3, [0,2], convention="left regular")
            sage: x = C([[5,4,1],[3,2,1,1]])
            sage: x.Epsilon()
            3*Lambda[1]
        """
        P = self.parent()
        WLR = P.weight_lattice_realization()
        La = WLR.fundamental_weights()
        n = self.normal_cells()
        return WLR.sum(len(n[i])*La[i] for i in P.index_set() if i in n)

    def Phi(self):
        r"""
        Return `\phi` of ``self``.

        EXAMPLES::

            sage: C = crystals.KleshchevPartitions(3, [0,2], convention="left regular")
            sage: x = C([[5,4,1],[3,2,1,1]])
            sage: x.Phi()
            3*Lambda[0] + 2*Lambda[1]
        """
        P = self.parent()
        WLR = P.weight_lattice_realization()
        La = WLR.fundamental_weights()
        c = self.conormal_cells()
        return WLR.sum(len(c[i])*La[i] for i in P.index_set() if i in c)

    def weight(self):
        r"""
        Return the weight of ``self``.

        EXAMPLES::

            sage: C = crystals.KleshchevPartitions(3, [0,2], convention="left regular")
            sage: x = C([[5,4,1], [3,2,1,1]])
            sage: x.weight()
            3*Lambda[0] - Lambda[1] - 5*delta
            sage: x.Phi() - x.Epsilon()
            3*Lambda[0] - Lambda[1]

            sage: C = crystals.KleshchevPartitions(3, [0,2], convention="right regular")
            sage: y = C([[5,1,1], [4,2,2,1,1]])
            sage: y.weight()
            6*Lambda[0] - 4*Lambda[1] - 4*delta
            sage: y.Phi() - y.Epsilon()
            6*Lambda[0] - 4*Lambda[1]

            sage: C = crystals.KleshchevPartitions(3, [0,2], convention="left regular")
            sage: y = C([[5,1,1], [4,2,2,1,1]])
            sage: y.weight()
            6*Lambda[0] - 4*Lambda[1] - 4*delta
            sage: y.Phi() - y.Epsilon()
            6*Lambda[0] - 4*Lambda[1]
        """
        WLR = self.parent().weight_lattice_realization()
        alpha = WLR.simple_roots()
        La = WLR.fundamental_weights()
        r = self.parent()._multicharge
        wt = WLR.sum(La[ZZ(x)] for x in r)
        return wt - WLR.sum(alpha[self.content(*c, multicharge=r)]
                            for c in self.cells())


class KleshchevPartitionCrystal(KleshchevPartition, KleshchevCrystalMixin):
    """
    Kleshchev partition with the crystal structure.
    """

    def e(self, i):
        r"""
        Return the action of `e_i` on ``self``.

        INPUT:

        - ``i`` -- an element of the index set

        EXAMPLES::

            sage: C = crystals.KleshchevPartitions(3, convention="left regular")
            sage: x = C([5,4,1])
            sage: x.e(0)
            sage: x.e(1)
            [5, 4]
        """
        P = self.parent()
        cell = self.good_cells(i)
        if cell is None:
            return None
        r, _ = cell
        mu = list(self)
        mu[r] -= 1
        return type(self)(P, mu)

    def f(self, i):
        r"""
        Return the action of `f_i` on ``self``.

        INPUT:

        - ``i`` -- an element of the index set

        EXAMPLES::

            sage: C = crystals.KleshchevPartitions(3, convention="left regular")
            sage: x = C([5,4,1])
            sage: x.f(0)
            [5, 5, 1]
            sage: x.f(1)
            sage: x.f(2)
            [5, 4, 2]
        """
        P = self.parent()
        cell = self.cogood_cells(i)
        if cell is None:
            return None
        r,c = cell
        mu = list(self)
        if c == 0:
            mu.append(1)
        else:
            mu[r] += 1
        return type(self)(P, mu)


class KleshchevPartitionTupleCrystal(KleshchevPartitionTuple, KleshchevCrystalMixin):
    """
    Kleshchev partition tuple with the crystal structure.
    """

    def e(self, i):
        r"""
        Return the action of `e_i` on ``self``.

        INPUT:

        - ``i`` -- an element of the index set

        EXAMPLES::

            sage: C = crystals.KleshchevPartitions(3, [0,2], convention="left regular")
            sage: x = C([[5,4,1],[3,2,1,1]])
            sage: x.e(0)
            sage: x.e(1)
            ([5, 4, 1], [2, 2, 1, 1])
        """
        P = self.parent()
        cell = self.good_cells(i)
        if cell is None:
            return None
        k, r, _ = cell
        mu = self.to_list()
        mu[k][r] -= 1
        return type(self)(P, mu)

    def f(self, i):
        r"""
        Return the action of `f_i` on ``self``.

        INPUT:

        - ``i`` -- an element of the index set

        EXAMPLES::

            sage: C = crystals.KleshchevPartitions(3, [0,2], convention="left regular")
            sage: x = C([[5,4,1],[3,2,1,1]])
            sage: x.f(0)
            ([5, 5, 1], [3, 2, 1, 1])
            sage: x.f(1)
            ([5, 4, 1], [3, 2, 2, 1])
            sage: x.f(2)
        """
        P = self.parent()
        cell = self.cogood_cells(i)
        if cell is None:
            return None
        k,r,c = cell
        mu = self.to_list()
        if c == 0:
            mu[k].append(1)
        else:
            mu[k][r] += 1
        return type(self)(P, mu)

#--------------------------------------------------
# Kleshchev partitions - parent classes
#--------------------------------------------------


class KleshchevPartitions(PartitionTuples):
    r"""
    Kleshchev partitions.

    A partition (tuple) `\mu` is Kleshchev if it can be recursively
    obtained by adding a sequence of good nodes to the empty
    :class:`PartitionTuple` of the same :meth:`~PartitionTuple.level`
    and multicharge.

    There are four different conventions that are used in the literature for
    Kleshchev partitions, depending on whether we read partitions from top
    to bottom (regular) or bottom to top (restricted) and whether we read
    partition tuples from left to right or right to left. All of these
    conventions are supported::

        sage: KleshchevPartitions(2, [0,0], size=2, convention='left regular')[:]
        [([1], [1]), ([2], [])]
        sage: KleshchevPartitions(2, [0,0], size=2, convention='left restricted')[:]
        [([1], [1]), ([], [1, 1])]
        sage: KleshchevPartitions(2, [0,0], size=2, convention='right regular')[:]
        [([1], [1]), ([], [2])]
        sage: KleshchevPartitions(2, [0,0], size=2, convention='right restricted')[:]
        [([1], [1]), ([1, 1], [])]

    By default, the ``left restricted`` convention is used. As a shorthand,
    ``LG``, ``LS``, ``RG`` and ``RS``, respectively, can be used to specify
    the ``convention``. With the ``left`` convention the partition tuples
    should be ordered with the most dominant partitions in the partition
    tuple on the left and with the ``right`` convention the most dominant
    partition is on the right.

    The :class:`~KleshchevPartitions` class will automatically convert
    between these four different conventions::

        sage: KPlg = KleshchevPartitions(2, [0,0], size=2, convention='left regular')
        sage: KPls = KleshchevPartitions(2, [0,0], size=2, convention='left restricted')
        sage: [KPlg(mu) for mu in KPls]
        [([1], [1]), ([2], [])]

    EXAMPLES::

        sage: sorted(KleshchevPartitions(5,[3,2,1],1, convention='RS'))
        [([], [], [1]), ([], [1], []), ([1], [], [])]
        sage: sorted(KleshchevPartitions(5, [3,2,1], 1, convention='LS'))
        [([], [], [1]), ([], [1], []), ([1], [], [])]
        sage: sorted(KleshchevPartitions(5, [3,2,1], 3))
        [([], [], [1, 1, 1]),
         ([], [], [2, 1]),
         ([], [], [3]),
         ([], [1], [1, 1]),
         ([], [1], [2]),
         ([], [1, 1], [1]),
         ([], [2], [1]),
         ([], [3], []),
         ([1], [], [1, 1]),
         ([1], [], [2]),
         ([1], [1], [1]),
         ([1], [2], []),
         ([1, 1], [1], []),
         ([2], [], [1]),
         ([2], [1], []),
         ([3], [], [])]
        sage: sorted(KleshchevPartitions(5, [3,2,1], 3, convention="left regular"))
        [([], [], [1, 1, 1]),
         ([], [1], [1, 1]),
         ([], [1], [2]),
         ([], [1, 1], [1]),
         ([], [1, 1, 1], []),
         ([1], [], [1, 1]),
         ([1], [1], [1]),
         ([1], [1, 1], []),
         ([1], [2], []),
         ([1, 1], [], [1]),
         ([1, 1], [1], []),
         ([1, 1, 1], [], []),
         ([2], [], [1]),
         ([2], [1], []),
         ([2, 1], [], []),
         ([3], [], [])]

    REFERENCES:

    - [AM2000]_
    - [Ariki2001]_
    - [BK2009]_
    - [Kle2009]_
    """
    @staticmethod
    def __classcall_private__(cls, e, multicharge=(0,), size=None,
                              convention="left restricted"):
        r"""
        This is a factory class which returns the appropriate parent based on
        the values of `level` and `size`.

        EXAMPLES::

            sage: sorted(KleshchevPartitions(5, [3,2,1], 1, convention='RS'))
            [([], [], [1]), ([], [1], []), ([1], [], [])]
            sage: sorted(KleshchevPartitions(5, [3,2,1], 1, convention='LS'))
            [([], [], [1]), ([], [1], []), ([1], [], [])]
        """
        if size is None and multicharge in ZZ:
            size = ZZ(multicharge)
            multicharge = (0,)

        I = IntegerModRing(e)
        multicharge = tuple([I(x) for x in multicharge])

        convention = convention.upper()
        if 'S' in convention:
            convention = convention[0] + 'S'
        elif 'G' in convention:
            convention = convention[0] + 'G'
        if convention not in ['RG','LG', 'RS', 'LS']:
            raise ValueError('invalid convention')

        if size is None:
            return KleshchevPartitions_all(e, multicharge, convention)

        return KleshchevPartitions_size(e, multicharge, size, convention)

    def multicharge(self):
        """
        Return the multicharge of ``self``.

        EXAMPLES::

            sage: KP = KleshchevPartitions(6, [2])
            sage: KP.multicharge()
            (2,)
            sage: KP = KleshchevPartitions(5, [3,0,1], 1, convention='LS')
            sage: KP.multicharge()
            (3, 0, 1)
        """
        return self._multicharge

    def convention(self):
        """
        Return the convention of ``self``.

        EXAMPLES::

            sage: KP = KleshchevPartitions(4)
            sage: KP.convention()
            'restricted'
            sage: KP = KleshchevPartitions(6, [4], 3, convention="right regular")
            sage: KP.convention()
            'regular'
            sage: KP = KleshchevPartitions(5, [3,0,1], 1)
            sage: KP.convention()
            'left restricted'
            sage: KP = KleshchevPartitions(5, [3,0,1], 1, convention='right regular')
            sage: KP.convention()
            'right regular'
        """
        if self._convention[1] == 'S':
            convention = "restricted"
        else:
            convention = "regular"

        if self._level == 1:
            return convention

        if self._convention[0] == 'R':
            return "right " + convention

        return "left " + convention

    def _element_constructor_(self, mu):
        r"""
        Return ``mu`` as an element of :class:`~KleshchevPartitions`, or
        or raise an error if ``mu`` is not a Kleshchev partition (tuple).

        The main purpose of the element constructor code is to allow automatic
        conversion between the four possible conventions for Kleshchev
        partitions.

        EXAMPLES::

            sage: KPlg = KleshchevPartitions(2, [0,0], size=2, convention='left regular')
            sage: KPls = KleshchevPartitions(2, [0,0], size=2, convention='left restricted')
            sage: [KPlg(mu) for mu in KPls] # indirect doc test
            [([1], [1]), ([2], [])]
        """
        if isinstance(mu, (KleshchevPartition, KleshchevPartitionTuple)):
            KPmu = mu.parent()
            if KPmu == self:
                return mu

            if KPmu._level != self._level or KPmu._e != self._e:
                raise ValueError('%s is not an element of %s' % (mu, self))

            if KPmu._convention[1] != self._convention[1]:
                mu = [nu.conjugate() for nu in mu]
                if self._level > 1 and KPmu._convention[0] == self._convention[0]:
                    mu = mu[::-1]

        return super()._element_constructor_(mu)


class KleshchevPartitions_all(KleshchevPartitions):
    r"""
    Class of all Kleshchev partitions.

    .. RUBRIC:: Crystal structure

    We consider type `A_{e-1}^{(1)}` crystals, and let `r = (r_i |
    r_i \in \ZZ / e \ZZ)` be a finite sequence of length `k`, which
    is the *level*, and `\lambda = \sum_i \Lambda_{r_i}`. We will
    model the highest weight `U_q(\mathfrak{g})`-crystal `B(\lambda)`
    by a particular subset of partition tuples of level `k`.

    Consider a partition tuple `\mu` with multicharge `r`.
    We define `e_i(\mu)` as the partition tuple obtained after the
    deletion of the `i`-:meth:`good cell
    <~sage.combinat.partition_kleshchev.KleshchevPartitionTuple.good_cell>`
    to `\mu` and `0` if there is no `i`-good cell. We define `f_i(\mu)` as
    the partition tuple obtained by the addition of the `i`-:meth:`cogood cell
    <~sage.combinat.partition_kleshchev.KleshchevPartitionTuple.cogood_cell>`
    to `\mu` and `0` if there is no `i`-good cell.

    The crystal `B(\lambda)` is the crystal generated by the empty
    partition tuple. We can compute the weight of an element `\mu` by taking
    `\lambda - \sum_{i=0}^n c_i \alpha_i` where `c_i` is the number of cells
    of `n`-residue `i` in `\mu`. Partition tuples in the crystal are known
    as *Kleshchev partitions*.

    .. NOTE::

        We can describe normal (not restricted) Kleshchev partition tuples
        in `B(\lambda)` as partition tuples `\mu` such that
        `\mu^{(t)}_{r_t - r_{t+1} + x} < \mu^{(t+1)}_x`
        for all `x \geq 1` and `1 \leq t \leq k - 1`.

    INPUT:

    - ``e`` -- for type `A_{e-1}^{(1)}` or `0`
    - ``multicharge`` -- the multicharge sequence `r`
    - ``convention`` -- (default: ``'LS'``) the reading convention

    EXAMPLES:

    We first do an example of a level 1 crystal::

        sage: C = crystals.KleshchevPartitions(3, [0], convention="left restricted")
        sage: C
        Kleshchev partitions with e=3
        sage: mg = C.highest_weight_vector()
        sage: mg
        []
        sage: mg.f(0)
        [1]
        sage: mg.f(1)
        sage: mg.f(2)
        sage: mg.f_string([0,2,1,0])
        [1, 1, 1, 1]
        sage: mg.f_string([0,1,2,0])
        [2, 2]
        sage: GC = C.subcrystal(max_depth=5).digraph()
        sage: B = crystals.LSPaths(['A',2,1], [1,0,0])
        sage: GB = B.subcrystal(max_depth=5).digraph()
        sage: GC.is_isomorphic(GB, edge_labels=True)
        True

    Now a higher level crystal::

        sage: C = crystals.KleshchevPartitions(3, [0,2], convention="right restricted")
        sage: mg = C.highest_weight_vector()
        sage: mg
        ([], [])
        sage: mg.f(0)
        ([1], [])
        sage: mg.f(2)
        ([], [1])
        sage: mg.f_string([0,1,2,0])
        ([2, 2], [])
        sage: mg.f_string([0,2,1,0])
        ([1, 1, 1, 1], [])
        sage: mg.f_string([2,0,1,0])
        ([2], [2])
        sage: GC = C.subcrystal(max_depth=5).digraph()
        sage: B = crystals.LSPaths(['A',2,1], [1,0,1])
        sage: GB = B.subcrystal(max_depth=5).digraph()
        sage: GC.is_isomorphic(GB, edge_labels=True)
        True

    The ordering of the residues gives a different representation of the
    higher level crystals (but it is still isomorphic)::

        sage: C2 = crystals.KleshchevPartitions(3, [2,0], convention="right restricted")
        sage: mg2 = C2.highest_weight_vector()
        sage: mg2.f_string([0,1,2,0])
        ([2], [2])
        sage: mg2.f_string([0,2,1,0])
        ([1, 1, 1], [1])
        sage: mg2.f_string([2,0,1,0])
        ([2, 1], [1])
        sage: GC2 = C2.subcrystal(max_depth=5).digraph()
        sage: GC.is_isomorphic(GC2, edge_labels=True)
        True

    TESTS:

    We check that all conventions give isomorphic crystals::

        sage: CLS = crystals.KleshchevPartitions(3, [2,0], convention="left restricted")
        sage: CRS = crystals.KleshchevPartitions(3, [2,0], convention="right restricted")
        sage: CLG = crystals.KleshchevPartitions(3, [2,0], convention="left regular")
        sage: CRG = crystals.KleshchevPartitions(3, [2,0], convention="right regular")
        sage: C = [CLS, CRS, CLG, CRG]
        sage: G = [B.subcrystal(max_depth=6).digraph() for B in C]
        sage: G[0].is_isomorphic(G[1], edge_labels=True)
        True
        sage: G[0].is_isomorphic(G[2], edge_labels=True)
        True
        sage: G[0].is_isomorphic(G[3], edge_labels=True)
        True

    REFERENCES:

    - [Ariki1996]_
    - [Ariki2001]_
    - [Tingley2007]_
    - [TingleyLN]_
    - [Vazirani2002]_
    """

    def __init__(self, e, multicharge, convention):
        r"""
        Initialize ``self``.

        EXAMPLES::

            sage: K = KleshchevPartitions(4, [2])
            sage: TestSuite(K).run()  # long time
            sage: K = KleshchevPartitions(4, [0,2,1])
            sage: TestSuite(K).run()  # long time

            sage: K = KleshchevPartitions(0, [2])
            sage: TestSuite(K).run()
            sage: K = KleshchevPartitions(0, [0,2,1])
            sage: TestSuite(K).run()  # long time
        """
        if e not in NN or e == 1:
            raise ValueError('e must belong to {0,2,3,4,5,6,...}')
        if e > 0:
            from sage.combinat.root_system.cartan_type import CartanType
            from sage.categories.highest_weight_crystals import HighestWeightCrystals
            from sage.categories.regular_crystals import RegularCrystals
            self._cartan_type = CartanType(['A', e-1, 1])
            cat = (HighestWeightCrystals(), RegularCrystals().Infinite())
        else:
            cat = InfiniteEnumeratedSets()

        self._level = len(multicharge)
        if self._level == 1:
            self.Element = KleshchevPartitionCrystal
            self._element_constructor_ = getattr_from_other_class(self, Partitions, '_element_constructor_')
        else:
            self.Element = KleshchevPartitionTupleCrystal

        super().__init__(category=cat)
        self._e = e   # for printing
        self._index_set = IntegerModRing(e)
        self._multicharge = multicharge
        self._convention = convention
        if e > 0:
            if self._level == 1:
                self.module_generators = (self.element_class(self, []),)
            else:
                self.module_generators = (self.element_class(self, [[]]*self._level),)

    def _repr_(self):
        """
        EXAMPLES::

            sage: KleshchevPartitions(4, [2])
            Kleshchev partitions with e=4
            sage: KleshchevPartitions(3,[0,0,0])
            Kleshchev partitions with e=3 and multicharge=(0,0,0)
            sage: KleshchevPartitions(3,[0,0,1])
            Kleshchev partitions with e=3 and multicharge=(0,0,1)
        """
        if self._level == 1:
            return 'Kleshchev partitions with e=%s' % (self._e)

        return 'Kleshchev partitions with e=%s and multicharge=(%s)' % (
                        self._e,','.join('%s' % m for m in self._multicharge))

    def __contains__(self, mu):
        """
        Containment test for Kleshchev partitions.

        EXAMPLES::

            sage: PartitionTuple([[3,2],[2]]) in KleshchevPartitions(2, [0,0], 7)
            False
            sage: PartitionTuple([[],[2,1],[3,2]]) in KleshchevPartitions(5, [0,0,1], 7)
            False
            sage: PartitionTuple([[],[2,1],[3,2]]) in KleshchevPartitions(5, [0,1,1], 7)
            False
            sage: PartitionTuple([[],[2,1],[3,2]]) in KleshchevPartitions(5, [0,1,1], 8)
            True
            sage: all(mu in PartitionTuples(3,8) for mu in KleshchevPartitions(2, [0,0,0], 8))
            True
        """
        if isinstance(mu, (KleshchevPartition, KleshchevPartitionTuple)):
            if mu.level() != self._level:
                return False
            mu = self.element_class(self, list(mu))
            if self._convention[1] == 'G':
                return mu.is_regular()

            return mu.is_restricted()

        try:
            mu = self.element_class(self, mu)
        except ValueError:
            return False
        return mu in self

    def __iter__(self):
        r"""
        Iterate over ``self``.

        EXAMPLES::

            sage: it = iter(KleshchevPartitions(2))
            sage: [next(it) for _ in range(10)]
            [[], [1], [1, 1], [2, 1], [1, 1, 1], [2, 1, 1],
             [1, 1, 1, 1], [2, 2, 1], [2, 1, 1, 1], [1, 1, 1, 1, 1]]
            sage: it = iter(KleshchevPartitions(2, convention='LG'))
            sage: [next(it) for _ in range(10)]
            [[], [1], [2], [3], [2, 1], [4], [3, 1], [5], [4, 1], [3, 2]]

            sage: it = iter(KleshchevPartitions(2, [0,1], convention='LS'))
            sage: [next(it) for _ in range(10)]
            [([], []),
             ([1], []),
             ([], [1]),
             ([1], [1]),
             ([], [1, 1]),
             ([1, 1], [1]),
             ([1], [1, 1]),
             ([], [2, 1]),
             ([], [1, 1, 1]),
             ([2, 1], [1])]
            sage: it = iter(KleshchevPartitions(2, [0,1], convention='RS'))
            sage: [next(it) for _ in range(10)]
            [([], []),
             ([1], []),
             ([], [1]),
             ([1, 1], []),
             ([1], [1]),
             ([2, 1], []),
             ([1, 1, 1], []),
             ([1, 1], [1]),
             ([1], [1, 1]),
             ([2, 1, 1], [])]
            sage: it = iter(KleshchevPartitions(2, [0,1], convention='LG'))
            sage: [next(it) for _ in range(10)]
            [([], []),
             ([1], []),
             ([], [1]),
             ([2], []),
             ([1], [1]),
             ([3], []),
             ([2, 1], []),
             ([2], [1]),
             ([1], [2]),
             ([4], [])]
            sage: it = iter(KleshchevPartitions(2, [0,1], convention='RG'))
            sage: [next(it) for _ in range(10)]
            [([], []),
             ([1], []),
             ([], [1]),
             ([1], [1]),
             ([], [2]),
             ([2], [1]),
             ([1], [2]),
             ([], [3]),
             ([], [2, 1]),
             ([2, 1], [1])]

            sage: it = iter(KleshchevPartitions(3, [0,1,2]))
            sage: [next(it) for _ in range(10)]
            [([], [], []), ([1], [], []), ([], [1], []), ([], [], [1]),
             ([1], [1], []), ([1], [], [1]), ([], [1, 1], []),
             ([], [1], [1]), ([], [], [2]), ([], [], [1, 1])]
        """
        # This is a modified form of what appears in the fixed size code
        if self._level == 1:
            if self._e == 0:
                P = Partitions()
            elif self._convention[1] == 'G':
                P = Partitions(regular=self._e)
            else:
                P = Partitions(restricted=self._e)

            for mu in P:
                yield self.element_class(self, list(mu))
        else:
            next_level = [self.element_class(self, [[]]*len(self._multicharge))]
            while True:
                cur = next_level
                next_level = []
                for mu in cur:
                    yield mu
                    mu_list = mu.to_list()
                    for cell in sorted(mu.cogood_cells().values()):
                        data = [list(p) for p in mu_list]
                        k, r, c = cell
                        if c == 0:
                            data[k].append(1)
                        else:
                            data[k][r] += 1
                        nu = self.element_class(self, data)
                        good_cells = nu.good_cells().values()
                        if self._convention[1] == "S":
                            if all(cell >= c for c in good_cells):
                                next_level.append(nu)
                        else:
                            if all(cell <= c for c in good_cells):
                                next_level.append(nu)

    def _an_element_(self):
        """
        Return a generic element.

        EXAMPLES::

            sage: KleshchevPartitions(3, [0,0,0,0], size=4).an_element()
            ([1], [1], [1], [1])
        """
        return self[12]


class KleshchevPartitions_size(KleshchevPartitions):
    """
    Kleshchev partitions of a fixed size.
    """

    def __init__(self, e, multicharge=(0,), size=0, convention='RS'):
        r"""
        Initialize ``self``.

        EXAMPLES::

            sage: K = KleshchevPartitions(4, 2)
            sage: TestSuite(K).run()
            sage: K = KleshchevPartitions(4, 4, convention='left regular')
            sage: TestSuite(K).run()
            sage: K = KleshchevPartitions(4, 4, convention='left restricted')
            sage: TestSuite(K).run()
            sage: K = KleshchevPartitions(4, [0,2,1], 4, convention='left regular')
            sage: TestSuite(K).run()
            sage: K = KleshchevPartitions(0, 2, convention='right restricted')
            sage: TestSuite(K).run()
            sage: K = KleshchevPartitions(0, [0,2,1], 4, convention='left restricted')
            sage: TestSuite(K).run()
            sage: K = KleshchevPartitions(0, [0,2,1], 4, convention='left regular')
            sage: TestSuite(K).run()

        We verify that we obtain the same size for all conventions
        and that the result is equal to the number of elements in
        the crystal at the corresponding depth::

            sage: B = crystals.LSPaths(['A',2,1], [1,0,1])
            sage: nd4 = (B.subcrystal(max_depth=4).cardinality()
            ....:        - B.subcrystal(max_depth=3).cardinality())
            sage: K = KleshchevPartitions(3, [0,2], 4, convention='RS')
            sage: K.cardinality() == nd4
            True
            sage: K = KleshchevPartitions(3, [0,2], 4, convention='RG')
            sage: K.cardinality() == nd4
            True
            sage: K = KleshchevPartitions(3, [0,2], 4, convention='LS')
            sage: K.cardinality() == nd4
            True
            sage: K = KleshchevPartitions(3, [0,2], 4, convention='LG')
            sage: K.cardinality() == nd4
            True
        """
        self._level = len(multicharge)
        if self._level == 1:
            self.Element = KleshchevPartition
            self._element_constructor_ = getattr_from_other_class(self, Partitions, '_element_constructor_')
        else:
            self.Element = KleshchevPartitionTuple
        super().__init__(category=FiniteEnumeratedSets())
        self._size = size
        # As lists do not take negative indices the case e=0 needs to be handled
        # differently. Rather than doing this we set e equal to a "really big"
        # number. Mathematically, this is equivalent and it means that we don't
        # have an exception to cater for.
        self._e = e
        self._I = IntegerModRing(e)
        self._multicharge = tuple(self._I(m) for m in multicharge)
        self._convention = convention

    def _repr_(self):
        """
        EXAMPLES::

            sage: KleshchevPartitions(4, [0,0], 3)
            Kleshchev partitions with e=4 and multicharge=(0,0) and size 3
        """
        if self._level == 1:
            return 'Kleshchev partitions with e=%s and size %s' % (self._e, self._size)

        return 'Kleshchev partitions with e=%s and multicharge=(%s) and size %s' % (
            self._e,','.join('%s' % m for m in self._multicharge), self._size
        )

    def __contains__(self, mu):
        """
        Check if ``mu`` is in ``self``.

        TESTS::

            sage: PartitionTuple([[3,2],[2]]) in KleshchevPartitions(2,[0,0],7)
            False
            sage: PartitionTuple([[3,2],[],[],[],[2]]) in KleshchevPartitions(5,[0,0,0,0,0],7)
            False
            sage: PartitionTuple([[2,1],[],[1,1],[],[2]]) in KleshchevPartitions(5,[0,0,0,0,0],7, convention='RG')
            False
            sage: PartitionTuple([[2,1],[],[1,1],[],[3]]) in KleshchevPartitions(2,[0,0,0,0,0],9, convention='RS')
            False
            sage: all(mu in PartitionTuples(3,8) for mu in KleshchevPartitions(0,[0,0,0],8))
            True
        """
        if isinstance(mu, (KleshchevPartition, KleshchevPartitionTuple)):
            if not (mu.level() == self._level and mu.size() == self._size):
                return False
            mu = self.element_class(self, list(mu))
            if self._convention[1] == 'G':
                return mu.is_regular()

            return mu.is_restricted()

        try:
            mu = self.element_class(self, mu)
        except ValueError:
            return False
        return mu in self

    def __iter__level_one(self):
        r"""
        Iterate over all Kleshchev partitions of level one and a fixed size.

        EXAMPLES::

            sage: KleshchevPartitions(2,0)[:]  # indirect doctest
            [[]]
            sage: KleshchevPartitions(2,1)[:]  # indirect doctest
            [[1]]
            sage: KleshchevPartitions(2,2)[:]  # indirect doctest
            [[1, 1]]
            sage: KleshchevPartitions(3,2)[:]  # indirect doctest
            [[2], [1, 1]]
        """
        if self._size == 0:
            yield self.element_class(self, [])
        else:
            if self._e == 0:
                P = Partitions(self._size)
            elif self._convention[1] == 'G':
                P = Partitions(self._size, regular=self._e)
            else:
                P = Partitions(self._size, restricted=self._e)

            for mu in P:
                yield self.element_class(self, list(mu))

    def __iter__higher_levels(self):
        r"""
        Iterate over all KleshchevPartitions of a fixed level greater than 1
        and a fixed size.

        EXAMPLES::

            sage: KleshchevPartitions(2,[0,0],1)[:]  # indirect doctest
            [([], [1])]
            sage: KleshchevPartitions(2,[0,0],2)[:]  # indirect doctest
            [([1], [1]), ([], [1, 1])]
            sage: KleshchevPartitions(3,[0,0],2)[:]  # indirect doctest
            [([1], [1]), ([], [2]), ([], [1, 1])]
        """
        if self._size == 0:
            yield self.element_class(self, [[]]*len(self._multicharge))
            return

        # For higher levels we have to recursively construct the restricted partitions
        # by adding on co-good nodes to smaller restricted partition. To avoid over
        # counting we return a new restricted partition only if we added on its lowest
        # good node.
        for mu in KleshchevPartitions_size(self._e, self._multicharge,
                                           size=self._size-1,
                                           convention=self._convention):
            mu_list = mu.to_list()
            for cell in mu.cogood_cells().values():
                data = [list(p) for p in mu_list]
                k,r,c = cell
                if c == 0:
                    data[k].append(1)
                else:
                    data[k][r] += 1
                nu = self.element_class(self, data)
                good_cells = nu.good_cells().values()
                if self._convention[1] == "S":
                    if all(cell >= c for c in good_cells):
                        yield nu
                else:
                    if all(cell <= c for c in good_cells):
                        yield nu

    @lazy_attribute
    def __iter__(self):
        """
        Wrapper to return the correct iterator which is different for
        :class:`Partitions` (level 1) and for :class:PartitionTuples`
        (higher levels).

        EXAMPLES::

            sage: KleshchevPartitions(3, 3, convention='RS')[:]
            [[2, 1], [1, 1, 1]]
            sage: KleshchevPartitions(3, 3, convention='RG')[:]
            [[3], [2, 1]]
            sage: KleshchevPartitions(3, [0], 3)[:]
            [[2, 1], [1, 1, 1]]
            sage: KleshchevPartitions(3, [0,0], 3)[:]
            [([1], [2]), ([1], [1, 1]), ([], [2, 1]), ([], [1, 1, 1])]
            sage: KleshchevPartitions(2, [0,1], size=0)[:]
            [([], [])]
            sage: KleshchevPartitions(2, [0,1], size=1)[:]
            [([1], []), ([], [1])]
            sage: KleshchevPartitions(2, [0,1], size=2)[:]
            [([1], [1]), ([], [1, 1])]
            sage: KleshchevPartitions(3, [0,1,2], size=2)[:]
            [([1], [1], []), ([1], [], [1]), ([], [1, 1], []),
             ([], [1], [1]), ([], [], [2]), ([], [], [1, 1])]
        """
        if self.level() == 1:
            return self.__iter__level_one

        return self.__iter__higher_levels

    def _an_element_(self):
        """
        Return a generic element.

        EXAMPLES::

            sage: KleshchevPartitions(4, [0,0,0,0], 4).an_element()
            ([1], [1], [1], [1])
            sage: KleshchevPartitions(4, [2], 4).an_element()
            [3, 1]
        """
        return self[0]

    Element = KleshchevPartitionTuple

#--------------------------------------------------
# helper functions
#--------------------------------------------------


def _a_good_cell(kpt, multicharge, convention):
    """
    Return a good cell from ``kpt`` considered as a Kleshchev partition
    with ``multicharge`` under ``convention``.

    EXAMPLES::

        sage: from sage.combinat.partition_kleshchev import _a_good_cell
        sage: I2 = IntegerModRing(2)
        sage: _a_good_cell([[3,2,1], [3,1,1]], [I2(0),I2(2)], 'RS')
        (1, 2, 0)
        sage: _a_good_cell([[3,1,1], [3,2]], [I2(0),I2(2)], 'LG')
        (1, 1, 1)
        sage: I3 = IntegerModRing(3)
        sage: _a_good_cell([[3,2,1], [3,1,1]], [I3(0),I3(2)], 'RS')
        (0, 2, 0)
        sage: _a_good_cell([[3,1,1], [3,2]], [I3(0),I3(2)], 'LG')
        (1, 0, 2)
        sage: _a_good_cell([[], []], [I3(0),I3(2)], 'RS') is None
        True
        sage: _a_good_cell([[1,1], []], [I3(0),I3(2)], 'LS') is None
        True
    """
    # We use a dictionary for the normal nodes as the indexing set is Z when e=0
    carry = defaultdict(int)        # a tally of #(removable cells)-#(addable cells)
    ret = None

    if convention[0] == 'L':
        rows = [(k,r) for k,part in enumerate(kpt) for r in range(len(part)+1)]
    else:
        rows = [(k,r) for k,part in reversed(list(enumerate(kpt))) for r in range(len(part)+1)]
    if convention[1] == 'S':
        rows.reverse()

    for row in rows:
        k,r = row
        if r == len(kpt[k]): # addable cell at bottom of a component
            carry[multicharge[k]-r] += 1
        else:
            res = multicharge[k] + kpt[k][r] - r - 1
            if r == len(kpt[k])-1 or kpt[k][r] > kpt[k][r+1]: # removable cell
                if carry[res] == 0:
                    ret = (k, r, kpt[k][r]-1)
                else:
                    carry[res] -= 1
            if r == 0 or kpt[k][r-1] > kpt[k][r]:             # addable cell
                carry[res+1] += 1

    # finally return the result
    return ret


def _is_regular(kpt, multicharge, convention):
    """
    Return ``True`` if ``kpt`` is a ``multicharge``-regular
    Kleshchev partition.

    EXAMPLES::

        sage: from sage.combinat.partition_kleshchev import _is_regular
        sage: I2 = IntegerModRing(2)
        sage: _is_regular([[3,1,1], [3,2]], [I2(0),I2(2)], 'LS')
        False
        sage: I3 = IntegerModRing(3)
        sage: _is_regular([[3,1,1], [3,2]], [I3(0),I3(2)], 'LS')
        True
        sage: _is_regular([[], []], [I3(0),I3(2)], 'LS')
        True
    """
    if all(part == [] for part in kpt):
        return True
    convention = convention[0] + 'G'
    cell = _a_good_cell(kpt, multicharge, convention)
    while cell is not None:
        k, r, _ = cell
        if kpt[k][r] == 1:
            kpt[k].pop()
        else:
            kpt[k][r] -= 1
        cell = _a_good_cell(kpt, multicharge, convention)
    return all(part == [] for part in kpt)


def _is_restricted(kpt, multicharge, convention):
    """
    Return ``True`` if ``kpt`` is an ``multicharge``-restricted
    Kleshchev partition.

    EXAMPLES::

        sage: from sage.combinat.partition_kleshchev import _is_restricted
        sage: I2 = IntegerModRing(2)
        sage: _is_restricted([[3,2,1], [3,1,1]], [I2(0),I2(2)], 'RG')
        False
        sage: I3 = IntegerModRing(3)
        sage: _is_restricted([[3,2,1], [3,1,1]], [I3(0),I3(2)], 'RG')
        True
        sage: _is_restricted([[], []], [I3(0),I3(2)], 'RG')
        True
    """
    if all(not part for part in kpt):
        return True
    convention = convention[0] + 'S'
    cell = _a_good_cell(kpt, multicharge, convention)
    while cell is not None:
        k, r, _ = cell
        if kpt[k][r] == 1:
            kpt[k].pop()
        else:
            kpt[k][r] -= 1
        cell = _a_good_cell(kpt, multicharge, convention)
    return all(not part for part in kpt)



# sage.doctest: optional - fricas
r"""
To add handling of a FriCAS type constructor, proceed as follows:

* in ``LazyParent``, add the translation of the constructor to a
  SageMath parent.

* in ``fricas.spad``, add an appropriate package, if necessary.

* in ``FRICAS_DOMAIN_DISPATCH``, map the type constructor to a function that
  constructs the package containing ``sexport`` and a function that
  constructs the element given the parsed string produced by
  ``sexport``.
"""
from sage.misc.cachefunc import cached_method
from sage.misc.lazy_import import lazy_import
from sage.rings.integer import Integer
lazy_import('sage.calculus.var', ['var', 'function'])
lazy_import('sage.symbolic.expression', ['symbol_table', 'register_symbol'])
lazy_import('sage.symbolic.constants', ['I', 'e', 'pi'])

FRICAS_CONSTANTS = {'%i': I,
                    '%e': e,
                    '%pi': pi}

# the dispatch dictionary for SEXPorter and Evaluator
# if the first element of the pair is a function, it is used by
# SEXPorter to rewrite the domain
FRICAS_DOMAIN_DISPATCH = {
    "Record": ("_record", "_eval_record"),
    "Union": ("_union", "_eval_union"),
    "String": ("_inputform", "_eval_call"),
    "Integer": ("_inputform", "_eval_call"),
    "PositiveInteger": ("_inputform", "_eval_call"),
    "NonNegativeInteger": ("_inputform", "_eval_call"),
    "Float": ("_inputform", "_eval_float"),
    "Boolean": ("_inputform", "_eval_bool"),
    "AlgebraicNumber": ("_inputform", "_eval_qqbar"),
    "Expression": ("_inputform", "_eval_sr"),
    "OrderedCompletion": ("_inputform", "_eval_sr"),
    "PiDomain": ("_inputform", "_eval_sr"),
    "PrimeField": ("_finite", "_eval_gf"),
    "IntegerMod": ("_finite", "_eval_call"),
    "FiniteField": ("_finite", "_eval_gf"),
    "Fraction": ("_unary", "_eval_fraction"),
    "List": ("_unary", "_eval_list"),
    "Matrix": ("_unary", "_eval_matrix"),
    "Polynomial": ("_unary", "_eval_polynomialring"),
    "Factored": ("_unary", "_eval_factorization"),
    "Vector": ("_unary", "_eval_list"),
    "DirectProduct": ("_aggregate", "_eval_list"),
    "UnivariatePolynomial": (lambda x: ("Polynomial", x[2]),
                             "_eval_polynomialring"),  # coerce to Polynomial
    "DistributedMultivariatePolynomial": (lambda x: ("Polynomial", x[2]),
                                          "_eval_polynomialring"),
    "MultivariatePolynomial": (lambda x: ("Polynomial", x[2]),
                               "_eval_polynomialring")}


class SEXPorter:
    """
    A class constructing the FriCAS package call to export
    objects in the given (parsed) domain.
    """
    def __init__(self, domain):
        """
        INPUT:

        - ``domain`` -- a nested list of strings and integers,
          describing a FriCAS domain.

        EXAMPLES::

            sage: from sage.interfaces.fricas_translator import SEXPorter
            sage: SEXPorter(("Integer",))._domain
            ('Integer',)
        """
        if (isinstance(domain, tuple)
            and (head := domain[0]) in FRICAS_DOMAIN_DISPATCH
            and callable(fun := FRICAS_DOMAIN_DISPATCH[head][0])):
            self._domain = fun(domain)
        else:
            self._domain = domain

    def _unparse(self):
        """
        Return the FriCAS domain as a string.

        .. WARNING::

            This method does not work with arguments that are lists,
            such as `MultivariatePolynomial([x,y], Integer)`.

        EXAMPLES::

            sage: from sage.interfaces.fricas_translator import SEXPorter
            sage: SEXPorter(("Fraction", ("FiniteField", 2, 3)))._unparse()
            'Fraction(FiniteField(2,3))'

            sage: SEXPorter(('Union', ('Fraction', ('Integer',)), '"failed"'))._unparse()
            'Union(Fraction(Integer),"failed")'

            sage: dom = ('Union', (':', 'f1', ('Expression', ('Integer',))), (':', 'f2', ('List', ('Expression', ('Integer',)))), (':', 'fail', '"failed"'), (':', 'pole', '"potentialPole"'))
            sage: SEXPorter(dom)._unparse()
            'Union(f1: Expression(Integer),f2: List(Expression(Integer)),fail: "failed",pole: "potentialPole")'
        """
        if not isinstance(self._domain, tuple):
            return str(self._domain)

        if len(self._domain) == 1:
            return str(self._domain[0])

        if self._domain[0] == ':' and len(self._domain) == 3:
            tag = self._domain[1]
            dom = SEXPorter(self._domain[2])._unparse()
            return f"{tag}: {dom}"

        return (self._domain[0] + "("
                + ",".join(SEXPorter(e)._unparse() for e in self._domain[1:])
                + ")")

    def _inputform(self):
        """
        Return an ``InputFormExport`` package call for this domain.

        EXAMPLES::

            sage: from sage.interfaces.fricas_translator import SEXPorter
            sage: SEXPorter(("Integer",))._inputform()
            '(sexport$InputFormExport(Integer))'
        """
        return f"(sexport$InputFormExport({self._unparse()}))"

    def _finite(self):
        """
        Return a ``FiniteExport`` package call for this domain.

        EXAMPLES::

            sage: from sage.interfaces.fricas_translator import SEXPorter
            sage: SEXPorter(("FiniteField", 2, 3))._finite()
            '(sexport$FiniteExport(FiniteField(2,3)))'
        """
        return f"(sexport$FiniteExport({self._unparse()}))"

    def _aggregate(self):
        """
        Return an ``AggregateExport`` package call for this domain.

        EXAMPLES::

            sage: from sage.interfaces.fricas_translator import SEXPorter
            sage: SEXPorter(('DirectProduct', 2, ('Integer',)))._aggregate()
            '(sexport$AggregateExport(Integer, DirectProduct(2,Integer), (sexport$InputFormExport(Integer))))'
        """
        inner = SEXPorter(self._domain[2])
        base_str = inner._unparse()
        export_str = inner.export_call()
        return f"(sexport$AggregateExport({base_str}, {self._unparse()}, {export_str}))"

    def _unary(self):
        """
        Return a ``UnaryExport`` package call for this domain.

        EXAMPLES::

            sage: from sage.interfaces.fricas_translator import SEXPorter
            sage: SEXPorter(("List", ("Integer",)))._unary()
            '(sexport$UnaryExport(Integer, (sexport$InputFormExport(Integer))))'
        """
        inner = SEXPorter(self._domain[1])
        base_str = inner._unparse()
        export_str = inner.export_call()
        return f"(sexport$UnaryExport({base_str}, {export_str}))"

    def _record(self):
        """
        Return a function call for a record.

        This constructs a function of the form::

            sexportrecordname1name2(obj) ==
                convert([sexport(obj.name1)$Domain1Export(...),
                         sexport(obj.name2)$Domain1Export(...)])@SExpression

        EXAMPLES::

            sage: from sage.interfaces.fricas_translator import SEXPorter
            sage: dom = ('Record', (':', 'particular', ('Integer',)), (':', 'basis', ('List', ('Integer',))))
            sage: SEXPorter(dom)._record()
            'sexportRecordparticularbasis'
        """
        from sage.interfaces.fricas import fricas

        def make_call(field_name, field_domain):
            return SEXPorter(field_domain).export_call() + f"(obj.{field_name})"

        items = ",".join(make_call(field[1], field[2]) for field in self._domain[1:])
        name = "sexportRecord" + "".join(field[1] for field in self._domain[1:])
        fricas.eval(f"{name}(obj) == convert([{items}])@SExpression")
        return name

    def _union(self):
        """
        Return a function call for a union.

        This constructs a function of the form::

            sexportunionname1name2(obj) == (
                obj case Domain1 => (Domain1 sexport(obj)$Domain1Export(...));
                obj case Domain2 => (Domain2 sexport(obj)$Domain2Export(...));
                (Domain3 sexport(obj)$Domain3Export(...)))

        EXAMPLES::

            sage: from sage.interfaces.fricas_translator import SEXPorter
            sage: dom = ('Union', ('Expression', ('Integer',)), ('PrimeField', 3))
            sage: SEXPorter(dom)._union()
            'sexportUnionExpressionIntegerPrimeField3'

            sage: dom = ('Union', ('Integer',), "failed")
            sage: SEXPorter(dom)._union()
            'sexportUnionIntegerfailed'
        """
        from sage.interfaces.fricas import fricas
        from sage.misc.flatten import flatten
        from re import compile

        def is_tagged(dom):
            return isinstance(dom, tuple) and len(dom) == 3 and dom[0] == ':'

        # detect if this is a tagged union (all fields tagged)
        tagged = all(is_tagged(dom) for dom in self._domain[1:])

        def make_call(field_domain, tag_index):
            if isinstance(field_domain, str):  # e.g., '"failed"'
                s = f'convert({field_domain})@SExpression'
            else:
                domain = SEXPorter(field_domain)
                s = f'{domain.export_call()}(obj)'

            return f'convert([convert({tag_index})@SExpression, {s}])@SExpression'

        def make_case(field, tag_index):
            if tagged:
                # field = (':', tag_name, domain)
                tag_name = field[1]
                field_domain = field[2]
            else:
                tag_name = None
                field_domain = field

            # for the last case we want no "case" prefix
            if tag_index == len(self._domain) - 2:
                return make_call(field_domain, tag_index)

            if tagged:
                return f'obj case {tag_name} => {make_call(field_domain, tag_index)}'

            domain = SEXPorter(field_domain)
            return f'obj case {domain._unparse()} => {make_call(field_domain, tag_index)}'

        items = "; ".join(make_case(dom, tag)
                          for tag, dom in enumerate(self._domain[1:]))

        name = "sexport" + "".join(map(str, flatten(self._domain)))
        pattern = compile(r'[\W_]+')
        name = pattern.sub('', name)

        fricas.eval(f"{name}(obj:{self._unparse()}): SExpression == ({items});")
        return name


    def export_call(self):
        """
        Return the function call doing the export, as a string.

        EXAMPLES::

            sage: from sage.interfaces.fricas_translator import SEXPorter
            sage: SEXPorter(("List", ("FiniteField", 2, 3))).export_call()
            '(sexport$UnaryExport(FiniteField(2,3), (sexport$FiniteExport(FiniteField(2,3)))))'

            sage: SEXPorter(("List", ("UnivariatePolynomial", "x", ("Integer",)))).export_call()
            '(sexport$UnaryExport(Polynomial(Integer), (sexport$UnaryExport(Integer, (sexport$InputFormExport(Integer))))))'

            sage: SEXPorter(('DirectProduct', 2, ('Integer',))).export_call()
            '(sexport$AggregateExport(Integer, DirectProduct(2,Integer), (sexport$InputFormExport(Integer))))'
        """
        if not isinstance(self._domain, tuple):
            # it should be a FriCAS string
            return self._domain

        head = self._domain[0]
        if head not in FRICAS_DOMAIN_DISPATCH:
            raise NotImplementedError(f"{head} cannot be translated from FriCAS to SageMath yet")

        return getattr(self, FRICAS_DOMAIN_DISPATCH[head][0])()


class SEXEvaluator:
    def __init__(self, ast, dom):
        """
        INPUT:

        - ``ast`` -- a nested list of strings and integers describing a FriCAS object
        - ``dom`` -- a :class:`LazyParent` describing the domain of ``ast``

        EXAMPLES::

            sage: from sage.interfaces.fricas_translator import SEXEvaluator, LazyParent
            sage: SEXEvaluator(1, LazyParent(['Integer']))._ast
            1
        """
        self._ast = ast
        self._dom = dom

    def eval(self):
        r"""
        Return the evaluation of ``self`` in the given parent.

        EXAMPLES::

            sage: from sage.interfaces.fricas_translator import SEXEvaluator, LazyParent
            sage: SEXEvaluator("5", LazyParent(['Integer'])).eval()
            5
        """
        head = self._dom.head()
        return getattr(self, FRICAS_DOMAIN_DISPATCH[head][1])()

    # specific evaluators
    # the naming convention is the parent in sage, in lower case

    def _eval_call(self):
        r"""
        Return the evaluation by passing it to the parent.

        EXAMPLES::

            sage: from sage.interfaces.fricas_translator import LazyParent, SEXEvaluator
            sage: SEXEvaluator("-5", LazyParent(['Integer']))._eval_call()
            -5
        """
        return self._dom.parent()(self._ast)

    def _eval_record(self):
        r"""
        Return the evaluation as a record.

        EXAMPLES::

            sage: from sage.interfaces.fricas_translator import LazyParent, SEXEvaluator
            sage: fricas._start()  # to fill the symbol table
            sage: obj = (1, (('exp', ('*', -1, 'x')),))
            sage: EXPR = ('Expression', ('Integer',))
            sage: dom = ('Record', (':', 'particular', EXPR), (':', 'basis', ('List', EXPR)))
            sage: result = SEXEvaluator(obj, LazyParent(dom)).eval()
            sage: result["particular"]
            1
            sage: result["basis"]
            [e^(-x)]
        """
        assert len(self._dom._domain) == len(self._ast) + 1
        result = dict()
        for field, value in zip(self._dom._domain[1:], self._ast):
            key = field[1]
            parent = LazyParent(field[2])
            val = SEXEvaluator(value, parent).eval()
            result[key] = val
        return result

    def _eval_union(self):
        r"""
        Return the evaluation of a union.

        This has to dispatch to the correct type

        EXAMPLES::

            sage: from sage.interfaces.fricas_translator import LazyParent, SEXEvaluator
            sage: obj = (1, (('exp', ('*', -1, 'x')),))
            sage: EXPR = ('Expression', ('Integer',))
            sage: dom = ('Record', (':', 'particular', EXPR), (':', 'basis', ('List', EXPR)))
            sage: result = SEXEvaluator(obj, LazyParent(dom)).eval()
            sage: result["particular"]
            1
            sage: result["basis"]
            [e^(-x)]
        """
        tag, val = self._ast
        dom = self._dom._domain[1 + tag]
        if isinstance(dom, tuple) and dom[0] == ':':
            dom = dom[2]
        if not isinstance(dom, tuple):
            assert dom == val
            return val[1:-1]
        parent = LazyParent(dom)
        return SEXEvaluator(val, parent).eval()

    def _eval_bool(self):
        r"""
        Return the evaluation as a boolean.

        EXAMPLES::

            sage: from sage.interfaces.fricas_translator import LazyParent, SEXEvaluator
            sage: SEXEvaluator("true", LazyParent(['Boolean']))._eval_bool()
            True
        """
        return self._ast == "true"

    def _eval_float(self):
        r"""
        Return the evaluation as an arbitrary precision float.

        EXAMPLES::

            sage: from sage.interfaces.fricas_translator import SEXParser, SEXPorter, SEXEvaluator, LazyParent
            sage: f = fricas("3.1415::Float")
            sage: P = f._check_valid()
            sage: dom_str = P.get_string(f"sageprint(dom({f._name})::Any)")
            sage: dom = SEXParser(dom_str).parse()
            sage: fun = SEXPorter(dom).export_call()
            sage: obj_str = P.get_string(f"sageprint({fun}({f._name}))")
            sage: obj = SEXParser(obj_str).parse(); obj
            ('float', 231801786030234225607, -66, 2)
            sage: SEXEvaluator(obj, LazyParent(dom))._eval_float()
            3.1415000000000000000
        """
        _, x, e, b = self._ast
        # Warning: precision$Float gives the current precision,
        # whereas the bitlength of two's complement gives the
        # precision of self.
        x = int(x)
        prec = max(x.bit_length() if x >= 0 else (~x).bit_length(), 53)
        P = self._dom.parent(prec=prec)
        return P(Integer(x) * Integer(b) ** Integer(e))

    def _eval_gf(self):
        r"""
        Return the evaluation as an element of a finite field.

        EXAMPLES::

            sage: from sage.interfaces.fricas_translator import SEXEvaluator, LazyParent
            sage: SEXEvaluator(5, LazyParent(['FiniteField', 2, 3]))._eval_gf()
            z3^2 + 1
        """
        P = self._dom.parent()
        return P.from_integer(self._ast % P.order())

    def _eval_list(self):
        r"""
        Return the evaluation by coercing a list to the parent.

        EXAMPLES::

            sage: from sage.interfaces.fricas_translator import SEXParser, LazyParent, SEXEvaluator
            sage: ast = SEXParser("(1 2 3)").parse()
            sage: dom = LazyParent(('List', ('Integer',)))
            sage: SEXEvaluator(ast, dom)._eval_list()
            [1, 2, 3]
        """
        base = self._dom.base()
        P = self._dom.parent()
        return P([SEXEvaluator(a, base).eval() for a in self._ast])

    def _eval_fraction(self):
        r"""
        Return the evaluation as a fraction.

        EXAMPLES::

            sage: from sage.interfaces.fricas_translator import SEXParser, LazyParent, SEXEvaluator
            sage: ast = SEXParser("(3 2)").parse()
            sage: SEXEvaluator(ast, LazyParent(('Fraction', ('Integer',))))._eval_fraction()
            3/2
        """
        base = self._dom.base()
        if not isinstance(self._ast, tuple):
            num = SEXEvaluator(self._ast, base).eval()
            return dom.parent()(num)

        a, b = self._ast
        num = SEXEvaluator(a, base).eval()
        den = SEXEvaluator(b, base).eval()
        return num / den

    def _eval_matrix(self):
        r"""
        Return the evaluation of ``ast`` as a matrix.

        EXAMPLES::

            sage: from sage.interfaces.fricas_translator import SEXParser, LazyParent, SEXEvaluator
            sage: ast = SEXParser("((1 2) (3 4))").parse()
            sage: SEXEvaluator(ast, LazyParent(('Matrix', ('Integer',))))._eval_matrix()
            [1 2]
            [3 4]

            sage: SEXEvaluator(ast, LazyParent(('Matrix', ('PrimeField', 3))))._eval_matrix()
            [1 2]
            [0 1]

            sage: ast = SEXParser("((1 2) (3 (sin x)))").parse()
            sage: SEXEvaluator(ast, LazyParent(('Matrix', ('Expression', ('Integer',)))))._eval_matrix()
            [     1      2]
            [     3 sin(x)]

            sage: ast = SEXParser("(((float 191846138366579336806 -67 2)))").parse()
            sage: SEXEvaluator(ast, LazyParent(('Matrix', ('Float',))))._eval_matrix()
            [1.3000000000000000000]
        """
        base = self._dom.base()
        m = [[SEXEvaluator(e, base).eval() for e in row]
             for row in self._ast]
        P = self._dom.parent()
        return P(m)

    def _eval_polynomialring(self):
        r"""
        Return the evaluation as a polynomial.

        EXAMPLES::

            sage: from sage.interfaces.fricas_translator import SEXParser, SEXPorter, SEXEvaluator, LazyParent
            sage: f = fricas("x^2*y - 3*z + 1")
            sage: P = f._check_valid()
            sage: dom_str = P.get_string(f"sageprint(dom({f._name})::Any)")
            sage: dom = SEXParser(dom_str).parse()
            sage: fun = SEXPorter(dom).export_call()
            sage: obj_str = P.get_string(f"sageprint({fun}({f._name}))")
            sage: obj = SEXParser(obj_str).parse(); obj
            (((('z', 1),), -3), ((('y', 1), ('x', 2)), 1), ((), 1))
            sage: SEXEvaluator(obj, LazyParent(dom))._eval_polynomialring()
            x^2*y - 3*z + 1
        """
        base = self._dom.base()
        names = tuple(sorted(set(v for mon, _ in self._ast for v, _ in mon)))
        P = self._dom.parent(names=names)

        from sage.rings.polynomial.polynomial_ring import PolynomialRing_generic
        if isinstance(P, PolynomialRing_generic):

            def to_exponent(mon):
                if len(mon):
                    return mon[0][1]
                return 0

            return P._from_dict({to_exponent(mon): SEXEvaluator(c, base).eval()
                                 for mon, c in self._ast})

        def to_tuple(mon):
            t = [0]*len(P._names)
            for v, e in mon:
                t[P._names.index(v)] = e
            return tuple(t)

        return P._from_dict({to_tuple(mon): SEXEvaluator(c, base).eval()
                             for mon, c in self._ast})

    def _eval_factorization(self):
        r"""
        Return the evaluation as a factored object.

        EXAMPLES::

            sage: from sage.interfaces.fricas_translator import SEXParser, SEXPorter, SEXEvaluator, LazyParent
            sage: f = fricas("-48").factor()
            sage: P = f._check_valid()
            sage: dom_str = P.get_string(f"sageprint(dom({f._name})::Any)")
            sage: dom = SEXParser(dom_str).parse()
            sage: fun = SEXPorter(dom).export_call()
            sage: obj_str = P.get_string(f"sageprint({fun}({f._name}))")
            sage: obj = SEXParser(obj_str).parse(); obj
            (-1, ((2, 4), (3, 1)))
            sage: SEXEvaluator(obj, LazyParent(dom))._eval_factorization()
            -1 * 2^4 * 3
        """
        from sage.structure.factorization import Factorization
        base = self._dom.base()
        unit, factors = self._ast
        return Factorization([(SEXEvaluator(f, base).eval(), e)
                              for f, e in factors],
                             unit=SEXEvaluator(unit, base).eval(),
                             sort=False,
                             simplify=False)

    def _eval_sr_aux(self):
        r"""
        Return the evaluation as an element of the
        symbolic ring.

        Post processing by :meth:`_eval_sr` is necessary.

        EXAMPLES:

        This function cannot use the symbol table to translate
        symbols which are not function calls, as :issue:`31849` shows
        ``D`` would erroneously be interpreted as differential
        then::

            sage: var("D")
            D
            sage: integrate(D/x, x, algorithm='fricas')
            D*log(x)

        However, it does have to check for constants, for example
        ``%pi``::

            sage: from sage.interfaces.fricas_translator import SEXEvaluator
            sage: SEXEvaluator("%pi", None)._eval_sr_aux()
            pi
        """
        from sage.symbolic.ring import SR

        if isinstance(self._ast, (Integer, float)):
            return SR(self._ast)

        if isinstance(self._ast, str):
            try:
                return FRICAS_CONSTANTS[self._ast]
            except KeyError:
                return var(self._ast.replace("%", "_"))

        f, *args = self._ast
        try:
            fun = symbol_table["fricas"][(f, len(args))]
        except KeyError:
            try:
                fun = symbol_table["fricas"][(f, -1)]
            except KeyError:
                fun = function(f)

        args_eval = [SEXEvaluator(a, self._dom)._eval_sr_aux() for a in args]
        return fun(*args_eval)

    def _eval_qqbar(self):
        r"""
        Evaluate as element of the algebraic field.

        EXAMPLES::

            sage: from sage.interfaces.fricas_translator import SEXParser, SEXPorter, SEXEvaluator, LazyParent
            sage: f = fricas("rootOf(x^5 - x - 1)$AN")
            sage: P = f._check_valid()
            sage: dom_str = P.get_string(f"sageprint(dom({f._name})::Any)")
            sage: dom = SEXParser(dom_str).parse()
            sage: fun = SEXPorter(dom).export_call()
            sage: obj_str = P.get_string(f"sageprint({fun}({f._name}))")
            sage: obj = SEXParser(obj_str).parse(); obj
            ('::',
             ('rootOf', ('+', ('+', ('^', 'x', 5), ('*', -1, 'x')), -1), 'x'),
             ('AlgebraicNumber',))
            sage: SEXEvaluator(obj, LazyParent(dom))._eval_qqbar()
            1.167303978261419?
        """
        ex = SEXEvaluator(self._ast[1], None)._eval_sr()
        return self._dom.parent()(ex)

    def _eval_sr(self):
        r"""
        Evaluate as an element of the symbolic ring.


        EXAMPLES::

            sage: from sage.interfaces.fricas_translator import SEXEvaluator
            sage: SEXEvaluator("x", None)._eval_sr()
            x
        """
        # a FriCAS expressions may contain implicit references to a
        # rootOf expression within itself, as for example in the
        # result of integrate(1/(1+x^5), x).  Each algebraic number
        # appearing in the expression is only introduced once and
        # assigned a variable (usually of the form %%...).
        rootOf = dict()  # (variable, polynomial)
        rootOf_ev = dict()  # variable -> (complex) algebraic number

        def convert_rootOf(x, y):
            if y in rootOf:
                assert rootOf[y] == x
            else:
                rootOf[y] = x
            return y

        register_symbol(convert_rootOf, {'fricas': 'rootOf'}, 2)
        ex = self._eval_sr_aux()

        # postprocessing of rootOf
        from sage.rings.qqbar import QQbar
        from sage.rings.polynomial.polynomial_ring_constructor import PolynomialRing
        while rootOf:
            for var, poly in rootOf.items():
                pvars = poly.variables()
                rvars = [v for v in pvars if v not in rootOf_ev]  # remaining variables
                uvars = [v for v in rvars if v in rootOf]  # variables to evaluate
                if len(uvars) == 1:
                    assert uvars[0] == var, "the only variable in uvars should be %s but is %s" % (var, uvars[0])
                    break
            else:
                assert False, "circular dependency in rootOf expression"
            # substitute known roots
            poly = poly.subs(rootOf_ev)
            evars = set(poly.variables()).difference([var])
            del rootOf[var]
            if evars:
                # we just need any root per FriCAS specification -
                # however, if there are extra variables, we cannot
                # use QQbar.any_root
                rootOf_ev[var] = poly.roots(var, multiplicities=False)[0]
            else:
                R = PolynomialRing(QQbar, "x")
                # PolynomialRing does not accept variable names with
                # leading underscores
                poly = R(poly.subs({var: R.gen()}))
                # we just need any root per FriCAS specification
                rootOf_ev[var] = poly.any_root()

        return ex.subs({var: (val.radical_expression()
                              if val.parent() is QQbar else val)
                        for var, val in rootOf_ev.items()})


class LazyParent:
    def __init__(self, domain):
        """
        Initialize this lazy parent from a FriCAS domain description.

        EXAMPLES::

            sage: from sage.interfaces.fricas_translator import LazyParent
            sage: LazyParent(['Integer']).head()
            'Integer'
        """
        self._domain = domain

    def head(self):
        """
        Return the constructor.

        EXAMPLES::

            sage: from sage.interfaces.fricas_translator import LazyParent
            sage: LazyParent(('List', ('Integer',))).head()
            'List'
        """
        return self._domain[0]

    def args(self):
        """
        Return the arguments of the constructor.

        EXAMPLES::

            sage: from sage.interfaces.fricas_translator import LazyParent
            sage: LazyParent(('List', ('Integer',))).args()
            (('Integer',),)
        """
        return self._domain[1:]

    @cached_method
    def base(self):
        """
        Return the base domain as a :class:`LazyParent`.

        EXAMPLES::

            sage: from sage.interfaces.fricas_translator import LazyParent
            sage: LazyParent(('List', ('Integer',))).base().head()
            'Integer'
        """
        head = self.head()
        args = self.args()
        if head in ["List",
                    "Fraction",
                    "Matrix",
                    "Polynomial",
                    "Factored",
                    "Vector"]:
            return LazyParent(args[0])

        if head in ["UnivariatePolynomial",
                    "MultivariatePolynomial",
                    "DistributedMultivariatePolynomial",
                    "DirectProduct"]:
            return LazyParent(args[1])

        raise NotImplementedError(f"Cannot build base for {head}")

    @cached_method
    def parent(self, **kwargs):
        """
        Return the corresponding SageMath parent.

        EXAMPLES::

            sage: from sage.interfaces.fricas_translator import LazyParent
            sage: LazyParent(['Integer']).parent()
            Integer Ring
        """
        head = self.head()
        args = self.args()

        if head == "String":
            return lambda x: x[1:-1]

        if head == "List":
            return list

        if head in ["Vector", "DirectProduct"]:
            from sage.modules.free_module_element import vector
            return vector

        if head in ["Matrix", "RectangularMatrix", "SquareMatrix"]:
            from sage.matrix.constructor import matrix
            return matrix

        if head in ["FiniteField", "PrimeField"]:
            from sage.rings.finite_rings.finite_field_constructor import GF
            return GF(*args)

        if head == "IntegerMod":
            from sage.rings.finite_rings.integer_mod_ring import IntegerModRing
            return IntegerModRing(*args)

        if head in ["Integer", "PositiveInteger", "NonNegativeInteger"]:
            from sage.rings.integer_ring import ZZ
            return ZZ

        if head == "AlgebraicNumber":
            from sage.rings.qqbar import QQbar
            return QQbar

        if head == "Expression":
            from sage.symbolic.ring import SR
            return SR

        if head == "Float":
            from sage.rings.real_mpfr import RealField
            return RealField(kwargs["prec"])

        if head == "Fraction":
            from sage.rings.fraction_field import FractionField
            base = self.base().parent(**kwargs)
            return FractionField(base)

        if head in ["UnivariatePolynomial",
                    "MultivariatePolynomial",
                    "DistributedMultivariatePolynomial"]:
            from sage.rings.polynomial.polynomial_ring_constructor import PolynomialRing
            base = self.base().parent(**kwargs)
            return PolynomialRing(base, args[0])

        if head == "Polynomial":
            from sage.rings.polynomial.polynomial_ring_constructor import PolynomialRing
            base = self.base().parent(**kwargs)
            if not hasattr(self, "_polynomial_symbols"):
                self._polynomial_symbols = []
            self._polynomial_symbols = sorted(set(list(kwargs.get("names", []))
                                                  + self._polynomial_symbols))
            # we always want a multivariate polynomial ring here
            return PolynomialRing(base,
                                  len(self._polynomial_symbols),
                                  names=self._polynomial_symbols)

        raise NotImplementedError(f"Cannot build parent for {head}")


class SEXParser:
    _WHITESPACE = " "
    _LEFTBRACKET = "("
    _RIGHTBRACKET = ")"
    _STRINGMARKER = '"'
    _ESCAPEMARKER = '_'  # STRINGMARKER must be escaped in strings

    def __init__(self, s):
        """
        INPUT:

        - ``s`` -- string

        EXAMPLES::

            sage: from sage.interfaces.fricas_translator import SEXParser
            sage: SEXParser("abc")._s
            'abc'
        """
        self._s = s
        self._start = 0  # specifies where to start parsing

    def parse(self):
        r"""
        Parse the string.

        TESTS::

            sage: from sage.interfaces.fricas_translator import SEXParser, SEXPorter
            sage: SEXParser("abc").parse()
            'abc'

            sage: SEXParser("(asin c)").parse()
            ('asin', 'c')

            sage: SEXParser("(pi)").parse()
            ('pi',)

            sage: f = fricas('(x + y/2)::DMP(["x", "y", "z"], FRAC INT)')
            sage: P = f._check_valid()
            sage: dom_str = P.get_string(f"sageprint(dom({f._name})::Any)")
            sage: dom = SEXParser(dom_str).parse(); dom
            ('DistributedMultivariatePolynomial',
             ('x', 'y', 'z'),
             ('Fraction', ('Integer',)))
        """
        a = self._start
        while self._s[a] in self._WHITESPACE:
            a += 1

        self._start = a
        if self._s[a] == self._LEFTBRACKET:
            return self._parse_list()
        if self._s[a] == self._STRINGMARKER:
            return self._parse_string()
        return self._parse_atom()

    def _parse_list(self):
        """
        Parse the initial part of a string, assuming that it is a
        whitespace separated list.

        TESTS::

            sage: from sage.interfaces.fricas_translator import SEXParser
            sage: SEXParser("()").parse()  # indirect doctest
            ()

            sage: SEXParser("(a b c)").parse()
            ('a', 'b', 'c')

            sage: SEXParser('(bcd)').parse()
            ('bcd',)
        """
        # self._start specifies the position of the left bracket
        assert self._s[self._start] == self._LEFTBRACKET
        self._start += 1

        args = []
        while self._s[self._start] != self._RIGHTBRACKET:
            e = self.parse()
            args.append(e)
            self._start += 1

        return tuple(args)

    def _parse_atom(self):
        """
        Parse the initial part of a string, assuming that it is an
        atom, but not a string.

        Symbols and numbers must not contain ``_WHITESPACE`` and
        ``_RIGHTBRACKET``.

        TESTS::

            sage: from sage.interfaces.fricas_translator import SEXParser
            sage: SEXParser("abc").parse()  # indirect doctest
            'abc'
            sage: SEXParser("123 xyz").parse()
            123
        """
        a = self._start
        b = len(self._s)

        while (a < b
               and self._s[a] not in self._WHITESPACE
               and self._s[a] != self._RIGHTBRACKET):
            a += 1

        token = self._s[self._start:a]
        self._start = a - 1
        try:
            return Integer(token)
        except TypeError:
            try:
                return float(token)
            except ValueError:
                return token

    def _parse_string(self):
        r"""
        Parse the initial part of a string, assuming that it represents a
        string.

        TESTS::

            sage: from sage.interfaces.fricas_translator import SEXParser
            sage: SEXParser('"abc" 123')._parse_string()
            '"abc"'

            sage: SEXParser('"" 123')._parse_string()
            '""'

            sage: SEXParser('"____" 123')._parse_string()
            '"__"'

            sage: SEXParser('"_a" 123')._parse_string()
            '"a"'

            sage: SEXParser('"_" _"" 123')._parse_string()
            '"" ""'

            sage: SEXParser('"(b c)"')._parse_string()
            '"(b c)"'
        """
        a = self._start  # the position of the left quote
        assert self._s[a] == self._STRINGMARKER
        b = a + 1
        a = b
        S = []

        while self._s[b] != self._STRINGMARKER:
            if self._s[b] == self._ESCAPEMARKER:
                S.append(self._s[a:b])
                b += 1
                a = b
            b += 1

        self._start = b
        S.append(self._s[a:b])
        return '"' + "".join(S) + '"'



"""
Root system data for type I
"""
#*****************************************************************************
#       Copyright (C) 2008-2009 Nicolas M. Thiery <nthiery at users.sf.net>,
#
#  Distributed under the terms of the GNU General Public License (GPL)
#                  http://www.gnu.org/licenses/
#*****************************************************************************

from .cartan_type import CartanType_standard_finite, CartanType_simple


class CartanType(CartanType_standard_finite, CartanType_simple):
    def __init__(self, n):
        """
        EXAMPLES::

            sage: ct = CartanType(['I',5])
            sage: ct
            ['I', 5]
            sage: ct._repr_(compact = True)
            'I5'
            sage: ct.rank()
            2
            sage: ct.index_set()
            (1, 2)

            sage: ct.is_irreducible()
            True
            sage: ct.is_finite()
            True
            sage: ct.is_affine()
            False
            sage: ct.is_crystallographic()
            False
            sage: ct.is_simply_laced()
            False

        TESTS::

            sage: TestSuite(ct).run()
        """
        assert n >= 1
        CartanType_standard_finite.__init__(self, "I", n)

    def _latex_(self):
        r"""
        Return a latex representation of ``self``.

        EXAMPLES::

            sage: latex(CartanType(['I',5]))
            I_2(5)
        """
        return "I_2({})".format(self.n)

    def rank(self):
        """
        Type `I_2(p)` is of rank 2.

        EXAMPLES::

            sage: CartanType(['I', 5]).rank()
            2
        """
        return 2

    def index_set(self):
        r"""
        Type `I_2(p)` is indexed by `\{1,2\}`.

        EXAMPLES::

            sage: CartanType(['I', 5]).index_set()
            (1, 2)
        """
        return (1, 2)

    def coxeter_diagram(self):
        """
        Return the Coxeter matrix for this type.

        EXAMPLES::

            sage: ct = CartanType(['I', 4])
            sage: ct.coxeter_diagram()                                                  # needs sage.graphs
            Graph on 2 vertices
            sage: ct.coxeter_diagram().edges(sort=True)                                 # needs sage.graphs
            [(1, 2, 4)]
            sage: ct.coxeter_matrix()                                                   # needs sage.graphs
            [1 4]
            [4 1]
        """
        from sage.graphs.graph import Graph
        return Graph([[1,2,self.n]], multiedges=False)

    def coxeter_number(self):
        """
        Return the Coxeter number associated with ``self``.

        EXAMPLES::

            sage: CartanType(['I',3]).coxeter_number()
            3
            sage: CartanType(['I',12]).coxeter_number()
            12
        """
        return self.n



r"""
Gammoids

Let `D` be a directed graph and let `E` and `T` be not necessarily disjoint
sets of vertices of `D`. Say a subset `X` of `E` is in a collection `I` if
`X` can be linked into a subset of `T`. This defines a gammoid `(E, I)`,
where `E` is the groundset of a matroid and `I` is its independent sets.

Some authors use a reverse convention, where instead of a set `T` of roots,
they have a starting set `S` that is linked into subsets of `E`. Here we use
the convention that the vertices `T` are at the end of the directed paths,
not the beginning.

To construct a gammoid, first import Gammoid from
:mod:`sage.matroids.gammoid`.
A digraph and a list of roots from the vertex set are required for input::

    sage: from sage.matroids.gammoid import *
    sage: edgelist = [(0,1),(1,2),(2,3),(3,4)]
    sage: D = DiGraph(edgelist)
    sage: M = Gammoid(D, roots=[4]); M
    Gammoid of rank 1 on 5 elements
    sage: N = Gammoid(D, roots=[4], groundset=range(1,5)); N
    Gammoid of rank 1 on 4 elements
    sage: M.delete(0) == N
    True
    sage: N.is_isomorphic(matroids.Uniform(1,4))
    True
    sage: O = Gammoid(D, roots=[3]); O
    Gammoid of rank 1 on 5 elements
    sage: O.rank([0])
    1
    sage: O.rank([4])
    0

AUTHORS:

- Zachary Gershkoff (2017-08-25): initial version
"""

# *****************************************************************************
#       Copyright (C) 2017 Zachary Gershkoff <zgersh2@lsu.edu>
#
#
#  Distributed under the terms of the GNU General Public License (GPL)
#  as published by the Free Software Foundation; either version 2 of
#  the License, or (at your option) any later version.
#                  https://www.gnu.org/licenses/
# *****************************************************************************

from sage.graphs.digraph import DiGraph
from sage.matroids.matroid import Matroid
from sage.matroids.minor_matroid import MinorMatroid


class Gammoid(Matroid):
    r"""
    The gammoid class.

    INPUT:

    - ``D`` -- a loopless digraph representing the gammoid
    - ``roots`` -- a subset of the vertices
    - ``groundset`` -- (optional) a subset of the vertices

    OUTPUT: :class:`Gammoid`; if ``groundset`` is not specified,
    the entire vertex set is used (and the gammoid will be strict)

    EXAMPLES::

        sage: from sage.matroids.gammoid import Gammoid
        sage: D = digraphs.TransitiveTournament(5)
        sage: M = Gammoid(D, roots=[3, 4]); M
        Gammoid of rank 2 on 5 elements
        sage: M.is_isomorphic(matroids.Uniform(2, 5))
        True
        sage: D.add_vertex(6)
        sage: N = Gammoid(D, roots=[3, 4])
        sage: N.loops()
        frozenset({6})
        sage: O = Gammoid(D, roots=[3, 4, 6])
        sage: O.coloops()
        frozenset({6})
        sage: O.full_rank()
        3
        sage: P = Gammoid(D, roots=[3, 4], groundset=[0, 2, 3]); P
        Gammoid of rank 2 on 3 elements
    """

    def __init__(self, D, roots, groundset=None):
        """
        See the class definition for full documentation.

        EXAMPLES::

            sage: from sage.matroids.gammoid import Gammoid
            sage: edgedict = {1: [2], 2: [3], 4: [1, 5], 5: [2, 3, 8],
            ....:             6: [4, 7], 7: [5, 8]}
            sage: D = DiGraph(edgedict)
            sage: M = Gammoid(D, roots=[1, 2, 3])
            sage: N1 = Gammoid(D, groundset=range(1, 8), roots=[1, 2, 3])
            sage: N2 = M.delete(8)
            sage: N1 == N2
            True

        TESTS::

            sage: from sage.matroids.gammoid import Gammoid
            sage: D = DiGraph([(0, 0), (0, 1), (1, 1)], loops=True)
            sage: M = Gammoid(D, roots=[1])
            Traceback (most recent call last):
            ...
            ValueError: cannot add edge from 0 to 0 in graph without loops
            sage: D = DiGraph([(0, 1), (0, 1)], multiedges=True)
            sage: M = Gammoid(D, roots=[1])
            sage: M.is_isomorphic(matroids.Uniform(1, 2))
            True
            sage: M = Gammoid(D, roots=[3])
            Traceback (most recent call last):
            ...
            ValueError: roots must be a subset of the vertices
            sage: M = Gammoid(D, roots=[1], groundset=[3])
            Traceback (most recent call last):
            ...
            ValueError: groundset must be a subset of the vertices

        ::

            sage: from sage.matroids.gammoid import Gammoid
            sage: edgedict = {1: [2], 2: [3], 4: [1, 5], 5: [2, 3, 8],
            ....:             6: [4, 7], 7: [5, 8]}
            sage: D = DiGraph(edgedict)
            sage: M = Gammoid(D, roots=[]); M
            Gammoid of rank 0 on 8 elements
            sage: M = Gammoid(D, roots=[], groundset=[]); M
            Gammoid of rank 0 on 0 elements

        ::

            sage: from sage.matroids.gammoid import Gammoid
            sage: M = Gammoid(digraphs.TransitiveTournament(5), roots=[3, 4])
            sage: TestSuite(M).run()
            sage: TestSuite(M).run(verbose=True)
            running ._test_category() . . . pass
            running ._test_new() . . . pass
            running ._test_not_implemented_methods() . . . pass
            running ._test_pickling() . . . pass
        """
        self._roots = frozenset(roots)
        vertices = frozenset(D.vertices())
        if not self._roots.issubset(vertices):
            raise ValueError("roots must be a subset of the vertices")

        if groundset is None:
            self._groundset = vertices
        else:
            self._groundset = frozenset(groundset)
            if not self._groundset.issubset(vertices):
                raise ValueError("groundset must be a subset of the vertices")

        # self._D is used for computations
        # self._G is used for referencing and hashing
        # loops will cause an error here, but multiedges will be ignored
        self._D = DiGraph([D.vertices(), D.edges(labels=False)])
        self._prune_vertices()
        self._G = self._D.copy(immutable=True)
        self._rootv = self._D.add_vertex()
        for v in roots:
            self._D.add_edge(v, self._rootv)

    def _prune_vertices(self):
        """
        Remove irrelevant vertices from the internal digraph.

        This will remove vertices that are not part of the groundset and
        cannot be used in a valid path between an element and a root.
        However, this will not remove a cycle of such vertices.

        EXAMPLES::

            sage: from sage.matroids.gammoid import Gammoid
            sage: D = digraphs.TransitiveTournament(5)
            sage: D.add_vertex(6)
            sage: M = Gammoid(D, roots=[3, 4], groundset=[0, 1, 4])
            sage: M.digraph().vertices()
            [0, 1, 2, 3, 4]
        """
        vertices_c = self._roots.union(self._groundset)
        extra_sources = set(self._D.sources()).difference(vertices_c)
        extra_sinks = set(self._D.sinks()).difference(vertices_c)
        while extra_sources or extra_sinks:
            self._D.delete_vertices(set(extra_sources).union(extra_sinks))
            extra_sources = set(self._D.sources()).difference(vertices_c)
            extra_sinks = set(self._D.sinks()).difference(vertices_c)

    def __hash__(self):
        r"""
        Return an invariant of the matroid.

        This function is called when matroids are added to a set. It is very
        desirable to override it so it can distinguish matroids on the same
        groundset, which is a very typical use case!

        .. WARNING::

            This method is linked to ``__richcmp__`` (in Cython) and ``__cmp__``
            or ``__eq__``/``__ne__`` (in Python). If you override one, you
            should (and, in Cython, \emph{must}) override the other!

        EXAMPLES::

            sage: from sage.matroids.gammoid import Gammoid
            sage: D = digraphs.TransitiveTournament(5)
            sage: M1 = Gammoid(D, roots=[3, 4])
            sage: M2 = Gammoid(D, roots=[2, 3])
            sage: hash(M1) == hash(M2)
            False
            sage: N1 = M1.delete(2)
            sage: N2 = Gammoid(D, roots=[3, 4], groundset=[0, 1, 3, 4])
            sage: hash(N1) == hash(N2)
            True
            sage: D.delete_edge(0, 4)
            sage: M3 = Gammoid(D, roots=[3, 4])
            sage: hash(M1) == hash(M3)
            False
        """
        return hash((self._G, self._groundset, self._roots))

    def __eq__(self, other):
        """
        Compare two matroids.

        For two graphic matroids to be equal, all attributes of the underlying
        graphs must be equal.

        INPUT:

        - ``other`` -- matroid

        OUTPUT: ``True`` if ``self`` and ``other`` have the same digraph,
        roots, and groundset; ``False`` otherwise

        EXAMPLES::

            sage: from sage.matroids.gammoid import Gammoid
            sage: D = digraphs.TransitiveTournament(5)
            sage: M1 = Gammoid(D, roots=[3, 4])
            sage: M2 = Gammoid(D, roots=[2, 3])
            sage: M1 == M2
            False
            sage: N1 = M1.delete(2)
            sage: N2 = Gammoid(D, roots=[3, 4], groundset=[0, 1, 3, 4])
            sage: N1 == N2
            True
        """
        if not isinstance(other, Gammoid):
            return False
        # The roots are implied by self._D
        return (self._D == other._D and self._groundset == other._groundset)

    def __ne__(self, other):
        """
        Compare two matroids.

        INPUT:

        - ``other`` -- matroid

        OUTPUT: ``False`` if ``self`` and ``other`` have the same digraphs,
        roots, and groundset; ``True`` otherwise

        EXAMPLES::

            sage: from sage.matroids.gammoid import Gammoid
            sage: edgedict = {1: [2], 2: [3], 4: [1, 5], 5: [2, 3, 8], 6: [4, 7],
            ....:             7: [5, 8]}
            sage: D = DiGraph(edgedict)
            sage: M = Gammoid(D, roots=[1, 2, 3])
            sage: D.relabel([1, 2, 3, 4, 5, 6, 7, 'a'])
            sage: N = Gammoid(D, roots=[1, 2, 3])
            sage: M == N
            False
            sage: M.delete(8) == N.delete('a')
            True
        """
        return (not self == other)

    def __reduce__(self):
        """
        Save the matroid for later reloading.

        OUTPUT:

        A tuple ``(unpickle, (version, data))``, where ``unpickle`` is the
        name of a function that, when called with ``(version, data)``,
        produces a matroid isomorphic to ``self``. ``version`` is an integer
        (currently 0) and ``data`` is a tuple ``(E, F, name)`` where ``E`` is
        the groundset, ``F`` is the dictionary of flats, and ``name`` is a
        custom name.

        EXAMPLES::

            sage: from sage.matroids.gammoid import Gammoid
            sage: edgelist = [(0, 4), (0, 5), (1, 0), (1, 4), (2, 0), (2, 1), (2, 3),
            ....:             (2, 6), (3, 4), (3, 5), (4, 0), (5, 2), (6, 5)]
            sage: D = DiGraph(edgelist)
            sage: M = Gammoid(D, roots=[4, 5, 6])
            sage: M.rename("Example Gammoid: " + repr(M))
            sage: M == loads(dumps(M)), loads(dumps(M))  # indirect doctest
            (True, Example Gammoid: Gammoid of rank 3 on 7 elements)
            sage: M.reset_name()
            sage: loads(dumps(M))
            Gammoid of rank 3 on 7 elements
        """
        import sage.matroids.unpickling
        data = (self._D, self._roots, self._groundset, self.get_custom_name())
        version = 0
        return sage.matroids.unpickling.unpickle_gammoid, (version, data)

    def digraph(self):
        """
        Return the digraph associated with the gammoid.

        EXAMPLES::

            sage: from sage.matroids.gammoid import Gammoid
            sage: edgelist = [(0, 4), (0, 5), (1, 0), (1, 4), (2, 0), (2, 1), (2, 3),
            ....:             (2, 6), (3, 4), (3, 5), (4, 0), (5, 2), (6, 5)]
            sage: D = DiGraph(edgelist)
            sage: M = Gammoid(D, roots=[4, 5, 6])
            sage: M.digraph() == D
            True
        """
        return self._G.copy(data_structure='sparse')

    def digraph_plot(self):
        """
        Plot the graph with color-coded vertices.

        Vertices that are elements but not roots will be shown as blue. Vertices that
        are roots but not elements are red. Vertices that are both are pink, and vertices
        that are neither are grey.

        EXAMPLES::

            sage: from sage.matroids.gammoid import Gammoid
            sage: D = digraphs.TransitiveTournament(4)
            sage: M = Gammoid(D, roots=[2, 3])
            sage: M.digraph_plot()
            Graphics object consisting of 11 graphics primitives
        """
        self._inter = frozenset(self._roots.intersection(self._groundset))
        self._buckets = frozenset(self._roots.difference(self._inter))
        self._ending = frozenset(self._groundset.difference(self._inter))
        self._therest = set(self._G.vertices()).difference(self._roots)
        self._therest = frozenset(self._therest.difference(self._groundset))
        # Vertices just in buckets set are red "#D55E00"
        # Vertices just in the ending set are blue "#0072B2"
        # Vertices in both are pink "#CC79A7"
        # Vertices in neither are grey "#999999"
        d = {"#D55E00": list(self._buckets), "#CC79A7": list(self._inter),
             "#0072B2": list(self._ending), "#999999": list(self._therest)}
        return self._G.plot(vertex_colors=d)

    def _rank(self, X):
        """
        Return the rank of a set.

        EXAMPLES::

            sage: from sage.matroids.gammoid import Gammoid
            sage: D = digraphs.TransitiveTournament(4)
            sage: M = Gammoid(D, roots=[2, 3])
            sage: M.rank([2])
            1
            sage: M.full_rank()
            2
            sage: M.rank(M.groundset())
            2
        """
        source = self._D.add_vertex()
        for x in X:
            self._D.add_edge(source, x)

        rank = len(self._D.vertex_disjoint_paths(source, self._rootv))
        self._D.delete_vertex(source)
        return rank

    def groundset(self):
        """
        Return the groundset of the matroid.

        EXAMPLES::

            sage: from sage.matroids.gammoid import Gammoid
            sage: edgelist = [(0, 4), (0, 5), (1, 0), (1, 4), (2, 0), (2, 1), (2, 3),
            ....:             (2, 6), (3, 4), (3, 5), (4, 0), (5, 2), (6, 5)]
            sage: D = DiGraph(edgelist)
            sage: M = Gammoid(D, roots=[4, 5, 6])
            sage: sorted(M.groundset())
            [0, 1, 2, 3, 4, 5, 6]
        """
        return self._groundset

    def _repr_(self):
        """
        Return a string representation of the matroid.

        EXAMPLES:

            sage: from sage.matroids.gammoid import Gammoid
            sage: edgelist = [(0, 4), (0, 5), (1, 0), (1, 4), (2, 0), (2, 1), (2, 3),
            ....:             (2, 6), (3, 4), (3, 5), (4, 0), (5, 2), (6, 5)]
            sage: D = DiGraph(edgelist)
            sage: M = Gammoid(D, roots=[4, 5, 6]); M
            Gammoid of rank 3 on 7 elements
        """
        self._mrank = str(self._rank(self._groundset))
        self._elts = str(len(self._groundset))
        return "Gammoid of rank " + self._mrank + " on " + self._elts + " elements"

    def _minor(self, contractions=frozenset(), deletions=frozenset()):
        """
        Return a minor.

        INPUT:

        - ``contractions`` -- frozenset; subset of ``self.groundset()`` to be contracted
        -  ``deletions`` -- frozenset; subset of ``self.groundset()`` to be deleted

        Assumptions: contractions are independent, deletions are coindependent,
        contractions and deletions are disjoint.

        OUTPUT:

        If there are contractions, a :class:`MinorMatroid`. If not, a :class:`Gammoid`.

        EXAMPLES::

            sage: from sage.matroids.gammoid import Gammoid
            sage: edgedict = {1: [2], 2: [3], 4: [1,5], 5: [2, 3, 8],
            ....:             6: [4, 7], 7: [5, 8]}
            sage: D = DiGraph(edgedict)
            sage: M = Gammoid(D, roots=[1, 2, 3]); M
            Gammoid of rank 3 on 8 elements
            sage: N1 = M.delete([3, 4])
            sage: N1.groundset()
            frozenset({1, 2, 5, 6, 7, 8})
        """
        # Unlike TransversalMatroid, we don't check for coloops to delete
        # because that would mean altering the roots
        if deletions:
            new_groundset = self.groundset().difference(deletions)
            N = Gammoid(self.digraph(), self._roots, new_groundset)
        else:
            N = self

        if contractions:
            return MinorMatroid(N, contractions=contractions, deletions=frozenset())
        return N

    def gammoid_extension(self, vertex, neighbors=[]):
        """
        Return a gammoid extended by an element.

        The new element can be a vertex of the digraph that is not in the starting set,
        or it can be a new source vertex.

        INPUT:

        - ``vertex`` -- a vertex of the gammoid's digraph that is not already in the
          groundset, or a new vertex
        - ``neighbors`` -- (optional) a set of vertices of the digraph

        OUTPUT:

        A :class:`Gammoid`. If ``vertex`` is not already in the graph, then the
        new vertex will be have edges to ``neighbors``. The new vertex will have
        in degree `0` regardless of whether or not ``neighbors`` is specified.

        EXAMPLES::

            sage: from sage.matroids.gammoid import Gammoid
            sage: edgedict = {1: [2], 2: [3], 4: [1, 5], 5: [2, 3, 8],
            ....:             6: [4, 7], 7: [5, 8]}
            sage: D = DiGraph(edgedict)
            sage: M = Gammoid(D, roots=[2, 3, 4], groundset=range(2, 9))
            sage: M1 = M.gammoid_extension(1)
            sage: M1.groundset()
            frozenset({1, 2, 3, 4, 5, 6, 7, 8})
            sage: N = Gammoid(D, roots=[2, 3, 4])
            sage: M1 == N
            True
            sage: M1.delete(1)
            Gammoid of rank 3 on 7 elements
            sage: M == M1.delete(1)
            True
            sage: M2 = M.gammoid_extension(9); sorted(M2.loops())
            [8, 9]
            sage: M4 = M.gammoid_extension(9, [1, 2, 3])
            sage: M4.digraph().neighbors_out(9)
            [1, 2, 3]
            sage: M4.digraph().neighbors_in(9)
            []

        TESTS::

            sage: from sage.matroids.gammoid import Gammoid
            sage: edgedict = {1: [2], 2: [3], 4: [1, 5], 5: [2, 3, 8],
            ....:             6: [4, 7], 7: [5, 8]}
            sage: D = DiGraph(edgedict)
            sage: M = Gammoid(D, roots=[2, 3, 4], groundset=range(2, 9))
            sage: M.gammoid_extension(2)
            Traceback (most recent call last):
            ...
            ValueError: cannot extend by element already in groundset
            sage: M.gammoid_extension(1, [2, 3, 4])
            Traceback (most recent call last):
            ...
            ValueError: neighbors of vertex in digraph cannot be changed
            sage: M.gammoid_extension(9, [9])
            Traceback (most recent call last):
            ...
            ValueError: neighbors must already be in graph
        """
        if vertex in self._groundset:
            raise ValueError("cannot extend by element already in groundset")
        elif vertex in self._G:
            if neighbors:
                raise ValueError("neighbors of vertex in digraph cannot be changed")
            new_groundset = set(self._groundset).union([vertex])
            return Gammoid(D=self._G, roots=self._roots,
                           groundset=new_groundset)
        else:
            if not set(neighbors).issubset(self._G.vertices()):
                raise ValueError("neighbors must already be in graph")
            D = self.digraph()
            D.add_vertex(vertex)
            for n in neighbors:
                D.add_edge(vertex, n)
            new_groundset = set(self._groundset).union([vertex])
            return Gammoid(D=D, roots=self._roots, groundset=new_groundset)

    def gammoid_extensions(self, vertices=None):
        """
        Return an iterator of Gammoid extensions.

        This will only consider extensions from vertices that are already present
        in the digraph.

        INPUT:

        - ``vertices`` -- (optional) a list of vertices not in the digraph

        OUTPUT:

        An iterator of Gammoids. If ``vertices`` is not specified, every vertex
        not already in the groundset will be considered.

        EXAMPLES::

            sage: from sage.matroids.gammoid import Gammoid
            sage: M = Gammoid(digraphs.TransitiveTournament(5), roots=[3, 4],
            ....:             groundset=[0, 1, 4])
            sage: [sorted(M1.groundset()) for M1 in M.gammoid_extensions()]
            [[0, 1, 2, 4], [0, 1, 3, 4]]
            sage: N = Gammoid(digraphs.TransitiveTournament(5), roots=[3, 4])
            sage: [sorted(N1.groundset()) for N1 in N.gammoid_extensions()]
            []

        ::

            sage: from sage.matroids.gammoid import Gammoid
            sage: edgelist =[(i, i+1) for i in range(10)]
            sage: M = Gammoid(DiGraph(edgelist), roots=[9], groundset=[0])
            sage: sum(1 for M1 in M.gammoid_extensions(vertices=[3, 4, 5]))
            3
            sage: sum(1 for M1 in M.gammoid_extensions())
            9
            sage: set([M1.is_isomorphic(matroids.Uniform(1, 2))
            ....:      for M1 in M.gammoid_extensions()])
            {True}
            sage: len([M1 for M1 in M.gammoid_extensions([0, 1, 2])])
            Traceback (most recent call last):
            ...
            ValueError: vertices must be in the digraph and not already in the groundset
        """
        free_vertices = set(self._G.vertices()).difference(self._groundset)
        if vertices is None:
            vertices = free_vertices
        else:
            vertices = set(vertices)
            if not vertices.issubset(free_vertices):
                raise ValueError("vertices must be in the digraph and not already in the groundset")
        for v in vertices:
            new_groundset = self._groundset.union([v])
            yield Gammoid(self._G, self._roots, new_groundset)
